"""Interpolation to a pixel grid.

There are two functions: one for interpolation of scalar fields, and one
for interpolation of vector fields.
"""

from typing import Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from ..snap import SnapLike
from ..snap.snap import get_array_from_input, get_array_in_code_units
from .splash import interpolate_cross_section, interpolate_projection

Extent = Tuple[float, float, float, float]


def interpolate(
    *,
    snap: SnapLike,
    quantity: Union[str, ndarray],
    x: Union[str, ndarray] = 'x',
    y: Union[str, ndarray] = 'y',
    z: Optional[Union[str, ndarray]] = None,
    interp: 'str',
    z_slice: Optional[float] = None,
    extent: Extent,
    **kwargs,
) -> ndarray:
    """Interpolate a quantity on the snapshot to a pixel grid.

    Parameters
    ----------
    snap
        The Snap (or SubSnap) object.
    quantity
        The quantity to visualize. Can be a string to pass to Snap, or
        a 1d array (N,) of scalar data, or a 2d array (N, 3) of
        vector data. If quantity is 2d, only the first two components
        are interpolated, i.e. quantity[:, 0] and quantity[:, 1].
        Default is None.
    x
        The x-coordinate for the visualization. Can be a string to
        pass to Snap, or a 1d array (N,). Default is 'x'.
    y
        The y-coordinate for the visualization. Can be a string to
        pass to Snap, or a 1d array (N,). Default is 'y'.
    z
        The z-coordinate for the visualization. Can be a string to
        pass to Snap, or a 1d array (N,). This is only required for
        cross-section plots. Default is 'z'.
    interp
        The interpolation type. Default is 'projection'.

        - 'projection' : 2d interpolation via projection to xy-plane
        - 'cross_section' : 3d interpolation via cross-section in
          z-direction
    z_slice
        The z-coordinate value of the cross-section slice. Default
        is 0.0.
    extent
        The xy extent of the image as (xmin, xmax, ymin, ymax).
    **kwargs
        Additional keyword arguments to pass to scalar_interpolation
        and vector_interpolation.

    Returns
    -------
    ndarray
        The interpolated quantity on a pixel grid as an ndarray.
    """
    quantity = get_array_from_input(snap, quantity)
    x = get_array_from_input(snap, x, 'x')
    y = get_array_from_input(snap, y, 'y')
    z = get_array_from_input(snap, z, 'z')
    h = get_array_in_code_units(snap, 'smooth')
    m = get_array_in_code_units(snap, 'mass')

    if interp == 'projection':
        cross_section = None
    elif interp == 'cross_section':
        if z_slice is None:
            z_slice = 0.0
        cross_section = z_slice

    if quantity.ndim == 1:
        interpolated_data = scalar_interpolation(
            quantity=quantity,
            x_coordinate=x,
            y_coordinate=y,
            z_coordinate=z,
            extent=extent,
            smoothing_length=h,
            particle_mass=m,
            hfact=snap.properties['hfact'],
            cross_section=cross_section,
            **kwargs,
        )

    elif quantity.ndim == 2:
        interpolated_data = vector_interpolation(
            quantity_x=quantity[:, 0],
            quantity_y=quantity[:, 1],
            x_coordinate=x,
            y_coordinate=y,
            z_coordinate=z,
            extent=extent,
            smoothing_length=h,
            particle_mass=m,
            hfact=snap.properties['hfact'],
            cross_section=cross_section,
            **kwargs,
        )

    else:
        raise ValueError('quantity.ndim > 2: cannot determine quantity')

    return interpolated_data


def scalar_interpolation(
    *,
    quantity: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    z_coordinate: Optional[ndarray] = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float] = (512, 512),
    cross_section: Optional[float] = None,
    density_weighted: Optional[bool] = None,
) -> ndarray:
    """Interpolate scalar quantity to a pixel grid.

    Parameters
    ----------
    quantity
        A scalar quantity on the particles to interpolate.
    x_coordinate
        Particle coordinate for x-axis in interpolation.
    y_coordinate
        Particle coordinate for y-axis in interpolation.
    z_coordinate
        Particle coordinate for z-axis. Only required for cross section
        interpolation.
    extent
        The range in the x- and y-direction as (xmin, xmax, ymin, ymax).
    smoothing_length
        The smoothing length on each particle.
    particle_mass
        The particle mass on each particle.
    hfact
        The smoothing length factor.
    number_of_pixels
        The pixel grid to interpolate the scalar quantity to, as
        (npixx, npixy).
    cross_section
        Cross section slice position as a z-value. If None, cross
        section interpolation is turned off. Default is off.
    density_weighted
        Use density weighted interpolation. Default is off.

    Returns
    -------
    ndarray
        An array of scalar quantities interpolated to a pixel grid with
        shape (npixx, npixy).
    """
    return _interpolate(
        quantity=quantity,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        z_coordinate=z_coordinate,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        cross_section=cross_section,
        density_weighted=density_weighted,
    )


def vector_interpolation(
    *,
    quantity_x: ndarray,
    quantity_y: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    z_coordinate: Optional[ndarray] = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float] = (512, 512),
    cross_section: Optional[float] = None,
    density_weighted: Optional[bool] = None,
) -> ndarray:
    """Interpolate scalar quantity to a pixel grid.

    Parameters
    ----------
    quantity_x
        The x-component of a vector quantity to interpolate.
    quantity_y
        The y-component of a vector quantity to interpolate.
    x_coordinate
        Particle coordinate for x-axis in interpolation.
    y_coordinate
        Particle coordinate for y-axis in interpolation.
    z_coordinate
        Particle coordinate for z-axis. Only required for cross section
        interpolation.
    extent
        The range in the x- and y-direction as (xmin, xmax, ymin, ymax).
    smoothing_length
        The smoothing length on each particle.
    particle_mass
        The particle mass on each particle.
    hfact
        The smoothing length factor.
    number_of_pixels
        The pixel grid to interpolate the scalar quantity to, as
        (npixx, npixy).
    cross_section
        Cross section slice position as a z-value. If None, cross
        section interpolation is turned off. Default is off.
    density_weighted
        Use density weighted interpolation. Default is off.

    Returns
    -------
    ndarray
        An array of vector quantities interpolated to a pixel grid with
        shape (2, npixx, npixy).
    """
    vecsmoothx = _interpolate(
        quantity=quantity_x,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        z_coordinate=z_coordinate,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        cross_section=cross_section,
        density_weighted=density_weighted,
    )
    vecsmoothy = _interpolate(
        quantity=quantity_y,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        z_coordinate=z_coordinate,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        cross_section=cross_section,
        density_weighted=density_weighted,
    )
    return np.stack((np.array(vecsmoothx), np.array(vecsmoothy)))


def _interpolate(
    *,
    quantity: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    z_coordinate: Optional[ndarray] = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float],
    cross_section: Optional[float] = None,
    density_weighted: Optional[bool] = None,
) -> ndarray:
    if cross_section is None:
        do_cross_section = False
    else:
        do_cross_section = True
        zslice = cross_section
        if z_coordinate is None:
            raise ValueError('Cross section interpolation requires z_coordinate')
    normalise = False
    if density_weighted is None:
        density_weighted = False
    if density_weighted:
        normalise = True

    npixx, npixy = number_of_pixels
    xmin, ymin = extent[0], extent[2]
    pixwidthx = (extent[1] - extent[0]) / npixx
    pixwidthy = (extent[3] - extent[2]) / npixy
    npart = len(smoothing_length)

    itype = np.ones(smoothing_length.shape)
    if density_weighted:
        weight = particle_mass / smoothing_length ** 3
    else:
        weight = hfact ** -3 * np.ones(smoothing_length.shape)

    if do_cross_section:
        interpolated_data = interpolate_cross_section(
            x=x_coordinate,
            y=y_coordinate,
            z=z_coordinate,
            hh=smoothing_length,
            weight=weight,
            dat=quantity,
            itype=itype,
            npart=npart,
            xmin=xmin,
            ymin=ymin,
            zslice=zslice,
            npixx=npixx,
            npixy=npixy,
            pixwidthx=pixwidthx,
            pixwidthy=pixwidthy,
            normalise=normalise,
        )
    else:
        interpolated_data = interpolate_projection(
            x=x_coordinate,
            y=y_coordinate,
            hh=smoothing_length,
            weight=weight,
            dat=quantity,
            itype=itype,
            npart=npart,
            xmin=xmin,
            ymin=ymin,
            npixx=npixx,
            npixy=npixy,
            pixwidthx=pixwidthx,
            pixwidthy=pixwidthy,
            normalise=normalise,
        )

    return interpolated_data



"""The utils sub-package."""

from .geometry import cartesian_to_polar
from .math import cross, norm
from .utils import get_extent_from_percentile, is_documented_by, time_string

__all__ = [
    'cartesian_to_polar',
    'cross',
    'is_documented_by',
    'get_extent_from_percentile',
    'norm',
    'time_string',
]



"""HDF5 files."""

from pathlib import Path
from typing import Union

import h5py


class HDF5File:
    """HDF5 files.

    Parameters
    ----------
    filename
        The path the HDF5 file.
    """

    def __init__(self, filename: Union[str, Path]):
        if isinstance(filename, str):
            path = Path(filename)
        else:
            path = filename

        self.file_path = path.expanduser().resolve()
        self.file_name = path.name
        self.file_extension = path.suffix[1:]
        self.file_type = 'HDF5'
        self.file_handle = self.open_file()

    def __repr__(self):
        """Repr dunder method."""
        return self.__str__()

    def __str__(self):
        """Str dunder method."""
        return str(self.file_handle)

    def open_file(self):
        """Open file."""
        if not self.file_path.is_file():
            raise FileNotFoundError('Cannot find snapshot file')
        return h5py.File(self.file_path, mode='r')

    def close_file(self):
        """Close file."""
        self.file_handle.close()



"""The Visualization class.

This class contains methods for visualizing smoothed particle
hydrodynamics simulation data.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from numpy import ndarray

from .. import Quantity
from .. import units as plonk_units
from ..snap import SnapLike
from ..snap.snap import get_array_from_input
from ..utils import get_extent_from_percentile, is_documented_by
from . import plots
from .interpolation import Extent, interpolate

_kind_to_object = {
    'render': 'image',
    'contour': 'contour',
    'quiver': 'quiver',
    'stream': 'streamplot',
}

_kind_to_function = {
    'render': plots.render_plot,
    'contour': plots.contour_plot,
    'quiver': plots.quiver_plot,
    'stream': plots.stream_plot,
}


class Visualization:
    """Visualize scalar and vector smoothed particle hydrodynamics data.

    Visualize SPH data as a particle plot, a rendered image, a contour
    plot, a vector plot, or a stream plot.

    Parameters
    ----------
    snap
        The associated Snap (or SubSnap) object to visualize.

    Attributes
    ----------
    snap
        The associated Snap (or SubSnap) object to visualize.
    fig
        The matplotlib Figure object of the plot.
    ax
        The matplotlib Axes object of the plot.
    units
        The units of the plot: 'quantity', 'extent', 'projection'. The
        values are pint Unit objects.
    extent
        A tuple (xmin, xmax, ymin, ymax) of the extent of the plot.
    objects
        A dictionary containing the matplotlib plot objects:

        - 'lines' : list of matplotlib Line2D objects for particle plots
        - 'image' : matplotlib AxesImage object for rendered plots
        - 'colorbar' : matplotlib Colorbar object for rendered plots
        - 'contour' : matplotlib QuadContourSet object for contour plots
        - 'quiver' : matplotlib Quiver object for quiver (arrow) plots
        - 'streamplot' : matplotlib StreamplotSet for stream plots
    data
        A dictionary containing the data interpolated to a pixel grid:
        'render', 'contour', 'quiver', 'stream'.
    """

    def __init__(self, snap: SnapLike):
        self.snap = snap
        self.fig: Any = None
        self.ax: Any = None
        self.units: Dict[str, Any] = {
            'quantity': None,
            'extent': None,
            'projection': None,
        }
        self.extent: Extent = (-1, -1, -1, -1)
        self.objects: Dict[str, Any] = {
            'lines': None,
            'image': None,
            'colorbar': None,
            'contour': None,
            'quiver': None,
            'streamplot': None,
        }
        self.data: Dict[str, ndarray] = {
            'render': None,
            'contour': None,
            'quiver': None,
            'stream': None,
        }

    def plot(
        self,
        *,
        quantity: Optional[Union[str, ndarray]] = None,
        x: Union[str, ndarray] = 'x',
        y: Union[str, ndarray] = 'y',
        z: Union[str, ndarray] = 'z',
        kind: Optional[str] = None,
        interp: str = 'projection',
        z_slice: float = 0.0,
        extent: Extent = (-1, -1, -1, -1),
        units=None,
        ax: Optional[Any] = None,
        **kwargs,
    ) -> Visualization:
        """Visualize smoothed particle hydrodynamics data.

        Visualize SPH data by showing the particle positions, a rendered
        image or a contour plot for scalar data, a quiver (arrow) plot
        or stream plot for vector data.

        Parameters
        ----------
        quantity
            The quantity to visualize. Can be a string to pass to Snap,
            or a 1d array (N,) of scalar data, or a 2d array (N, 3) of
            vector data. If quantity is 2d, only the first two
            components are visualized, i.e. quantity[:, 0] and
            quantity[:, 1]. Default is None.
        x
            The x-coordinate for the visualization. Can be a string to
            pass to Snap, or a 1d array (N,). Default is 'x'.
        y
            The y-coordinate for the visualization. Can be a string to
            pass to Snap, or a 1d array (N,). Default is 'y'.
        z
            The z-coordinate for the visualization. Can be a string to
            pass to Snap, or a 1d array (N,). This is only required for
            cross-section plots. Default is 'z'.
        kind
            The type of plot.

            - 'particle' : particle plot (default if quantity is None)
            - 'render' : rendered image (default for scalar quantities)
            - 'contour' : contour plot (scalar quantity)
            - 'quiver' : quiver (arrow) plot (default for vector
              quantities)
            - 'stream' : stream plot (vector quantity)
        interp
            The interpolation type. Default is 'projection'.

            - 'projection' : 2d interpolation via projection to xy-plane
            - 'cross_section' : 3d interpolation via cross-section in
              z-direction
        z_slice
            The z-coordinate value of the cross-section slice. Default
            is 0.0.
        extent
            The range in the x and y-coord as (xmin, xmax, ymin, ymax).
            The default is to set the extent to a box of size set by an
            inscribed sphere containing 99% of particles.
        units
            The units of the plot as a dictionary with keys 'quantity',
            'extent', 'projection'. The values are Pint Unit objects.
        ax
            A matplotlib Axes handle.
        **kwargs
            Additional keyword arguments to pass to interpolation and
            matplotlib functions.

        Notes
        -----
        Additional parameters passed as keyword arguments will be
        passed to lower level functions as required. E.g. Plonk uses
        matplotlib's imshow for a render plot, so additional arguments
        to imshow can be passed this way.

        See below for additional parameters for interpolation,
        colorbars, quiver plots, etc. All other keyword arguments are
        passed to the appropriate matplotlib function.

        Other Parameters
        ----------------
        number_of_pixels : tuple
            The number of pixels to interpolate particle quantities
            to as a tuple (nx, ny). Default is (512, 512).
        density_weighted : bool
            Whether to density weight the interpolation or not.
            Default is False.
        fmt : str
            This is the matplotlib ax.plot method positional
            argument format string. Default is 'k.'.
        show_colorbar : bool
            Whether or not to display a colorbar. Default is True.
        number_of_arrows : tuple
            The number of arrows to display by sub-sampling the
            interpolated data. Default is (25, 25).
        normalize_vectors : bool
            Whether to normalize the arrows to all have the same
            length. Default is False.

        Examples
        --------
        Plot the particles.

        >>> viz = plonk.visualize.plot(snap=snap)

        Render the surface density in xy-plane.

        >>> viz = plonk.visualize.plot(
        ...     snap=snap,
        ...     quantity='density',
        ... )

        Get the interpolation to grid directly (without plotting).

        >>> grid_data = plonk.visualize.interpolate(
        ...     snap=snap,
        ...     quantity='density',
        ...     interp='projection',
        ...     extent=(-100, 100, -100, 100),
        ... )

        Make an animation of multiple snaps.

        >>> plonk.visualize.animation(
        ...     snaps=snaps,
        ...     quantity='density',
        ...     extent=(-100, 100, -100, 100),
        ...     vmin=0.0,
        ...     vmax=1.0,
        ...     filename='animation.mp4',
        ... )

        Set units for the plot.

        >>> units = {
        ...     'quantity': plonk.units('g / cm ** 3'),
        ...     'extent': plonk.units('au'),
        ...     'projection': plonk.units('cm'),
        ... }

        >>> viz = plonk.visualize.plot(
        ...     snap=snap,
        ...     quantity='density',
        ...     units=units,
        ... )

        You can also use the utility function to generate the units
        dictionary.

        >>> units = plonk.visualize.str_to_units('g/cm^3', 'au', 'cm')
        """
        if self.ax is None:
            if ax is None:
                self.fig, self.ax = plt.subplots()
            else:
                self.fig = ax.get_figure()
                self.ax = ax
        else:
            if ax is not None:
                raise ValueError('Trying to change existing Axes attribute')

        if extent == (-1, -1, -1, -1):
            extent = get_extent_from_percentile(self.snap)

        quantity_str: Optional[str] = quantity if isinstance(quantity, str) else None
        quantity, x, y, z, kind = _check_input(
            snap=self.snap, quantity=quantity, x=x, y=y, z=z, kind=kind
        )

        if quantity is not None:
            interpolation_kwargs = ('number_of_pixels', 'density_weighted')
            _kwargs = {
                key: val for key, val in kwargs.items() if key in interpolation_kwargs
            }
            for key in _kwargs:
                kwargs.pop(key)
            interpolated_data = interpolate(
                snap=self.snap,
                quantity=quantity,
                x=x,
                y=y,
                z=z,
                interp=interp,
                z_slice=z_slice,
                extent=extent,
                **_kwargs,
            )
            if units is not None:
                if quantity_str is None:
                    raise ValueError(
                        'Cannot set units when passing in arrays. '
                        'Instead, use strings to access\n'
                        'quantities on Snap. E.g. plot(..., quantity="density", ...).'
                    )
                interpolated_data, extent = _convert_units(
                    snap=self.snap,
                    quantity_str=quantity_str,
                    interpolated_data=interpolated_data,
                    extent=extent,
                    units=units,
                    interp=interp,
                )
                self.units.update(units)

        self.extent = extent

        show_colorbar = kwargs.pop('show_colorbar', True if kind == 'render' else False)

        if kind == 'particle':
            self.objects['lines'] = plots.particle_plot(
                snap=self.snap, x=x, y=y, extent=extent, ax=self.ax, **kwargs,
            )

        elif kind in ('render', 'contour', 'quiver', 'stream'):
            self.data[kind] = interpolated_data
            self.objects[_kind_to_object[kind]] = _kind_to_function[kind](
                interpolated_data=interpolated_data,
                extent=extent,
                ax=self.ax,
                **kwargs,
            )

        else:
            raise ValueError('Cannot determine plot type')

        if show_colorbar:
            divider = make_axes_locatable(self.ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            self.objects['colorbar'] = self.fig.colorbar(self.objects['image'], cax)

        self.ax.set_xlim(*extent[:2])
        self.ax.set_ylim(*extent[2:])
        self.ax.set_aspect('equal')

        return self

    def __repr__(self):
        """Dunder repr method."""
        return '<plonk.Visualization>'


class MultiVisualization:
    """Visualize multiple snaps.

    Attributes
    ----------
    snaps
        A list of snaps.
    options
        A dictionary of arguments passed to Visualization plot method.
    visualization
        The Visualization object.
    ax
        The matplotlib Axes object.

    Parameters
    ----------
    snaps
        A list of Snap objects.
    **kwargs
        Keyword arguments to pass to Visualization plot method.
    """

    def __init__(self, snaps, **kwargs):
        self.snaps = snaps
        self.options = kwargs
        self.visualization = Visualization(snap=snaps[0]).plot(**kwargs)
        self.ax = self.visualization.ax

        self._len = -1
        self._where = 0

    def _fn(self, idx):
        self.ax.clear()
        cbar = self.visualization.objects['colorbar']
        if cbar is not None:
            cbar.remove()
        viz = Visualization(snap=self.snaps[idx]).plot(ax=self.ax, **self.options)
        self.visualization = viz
        return viz

    def next(self, number: int = 1):
        """Visualize next snap."""
        idx = self._where + number
        if idx < len(self):
            self.visualization = self._fn(idx)
            self._where += number
        else:
            print('Too far forward. Going to last snap.')
            self.visualization = self._fn(len(self) - 1)
            self._where = len(self) - 1

    def prev(self, number: int = 1):
        """Visualize previous snap."""
        idx = self._where - number
        if idx > 0:
            self.visualization = self._fn(idx)
            self._where -= number
        else:
            print('Too far back. Going to first snap.')
            self.visualization = self._fn(0)
            self._where = 0

    def goto(self, idx):
        """Visualize particular snap by index."""
        if -len(self) < idx < len(self) - 1:
            self.visualization = self._fn(idx)
            self._where = np.mod(idx, len(self))
        else:
            raise ValueError('out of range')

    @property
    def index(self):
        """Current snap index."""
        return self._where

    def __len__(self):
        """Length as number of snaps."""
        if self._len == -1:
            self._len = len(self.snaps)
        return self._len


def _check_input(*, snap, quantity, x, y, z, kind):

    if quantity is not None:
        quantity = get_array_from_input(snap, quantity)

    x = get_array_from_input(snap, x, 'x')
    y = get_array_from_input(snap, y, 'y')
    z = get_array_from_input(snap, z, 'z')

    if quantity is not None:
        if quantity.ndim > 2:
            raise ValueError('Cannot interpret quantity')
        if kind in ('render', 'contour') and quantity.ndim != 1:
            raise ValueError('quantity is wrong shape for render or contour')
        if kind in ('quiver', 'stream') and quantity.ndim != 2:
            raise ValueError('quantity is wrong shape for quiver or streamplot')
        if kind is None:
            if quantity.ndim == 1:
                kind = 'render'
            elif quantity.ndim == 2:
                kind = 'quiver'
    else:
        if kind is None:
            kind = 'particle'
        else:
            raise ValueError(f'No quantity: can only do particle plot')

    return quantity, x, y, z, kind


def _convert_units(
    *,
    snap: SnapLike,
    quantity_str: str,
    interpolated_data: ndarray,
    extent: Extent,
    units: Dict[str, Any],
    interp: str,
):
    required_keys = {'extent', 'projection', 'quantity'}
    if not set(units) == required_keys:
        raise ValueError(f'units dictionary requires: {required_keys}')
    _quantity_unit = snap.get_array_unit(quantity_str)
    if interp == 'projection':
        data = (
            (interpolated_data * _quantity_unit * snap.units['length'])
            .to(units['quantity'] * units['projection'])
            .magnitude
        )
    elif interp == 'cross_section':
        data = (interpolated_data * _quantity_unit).to(units['quantity']).magnitude

    new_extent = tuple((extent * snap.units['length']).to(units['extent']).magnitude)

    return data, new_extent


@is_documented_by(Visualization.plot)
def plot(
    *,
    snap: SnapLike,
    quantity: Optional[Union[str, ndarray]] = None,
    x: Union[str, ndarray] = 'x',
    y: Union[str, ndarray] = 'y',
    z: Union[str, ndarray] = 'z',
    kind: Optional[str] = None,
    interp: str = 'projection',
    z_slice: float = 0.0,
    extent: Extent = (-1, -1, -1, -1),
    ax: Optional[Any] = None,
    **kwargs,
) -> Visualization:
    viz = Visualization(snap)
    viz.plot(
        quantity=quantity,
        x=x,
        y=y,
        z=z,
        kind=kind,
        interp=interp,
        z_slice=z_slice,
        extent=extent,
        ax=ax,
        **kwargs,
    )
    return viz


def plot_snaps(snaps: List[SnapLike], **kwargs) -> MultiVisualization:
    """Visualize multiple snaps.

    Parameters
    ----------
    snaps
        A list of Snap objects.
    **kwargs
        Keyword arguments to pass to Visualization plot method.

    Returns
    -------
    MultiVisualization

    Examples
    --------
    Initialize object passing in plotting parameters.

    >>> vi = plot_snaps(
    ...     snaps=sim.snaps,
    ...     quantity='density',
    ... )

    Go forwards and backwards through snaps.

    >>> vi.next()
    >>> vi.prev()

    Go to a particular snap, or skip ahead.

    >>> vi.goto(10)
    >>> vi.next(5)
    """
    return MultiVisualization(snaps, **kwargs)


def str_to_units(quantity, extent, projection):
    """Convert string to plonk units.

    Parameters
    ----------
    quantity
        The units string for the quantity.
    extent
        The units string for the plot extent.
    projection
        The units string for projection interpolation.
    """
    if isinstance(quantity, str):
        quantity = plonk_units(quantity)
    elif isinstance(quantity, Quantity):
        pass
    else:
        raise ValueError(f'Cannot determine quantity unit')
    if isinstance(extent, str):
        extent = plonk_units(extent)
    elif isinstance(extent, Quantity):
        pass
    else:
        raise ValueError(f'Cannot determine extent unit')
    if isinstance(projection, str):
        projection = plonk_units(projection)
    elif isinstance(projection, Quantity):
        pass
    else:
        raise ValueError(f'Cannot determine projection unit')

    if extent.dimensionality != plonk_units('cm').dimensionality:
        raise ValueError('extent has incorrect dimensions')
    if projection.dimensionality != plonk_units('cm').dimensionality:
        raise ValueError('projection has incorrect dimensions')

    return {'quantity': quantity, 'extent': extent, 'projection': projection}



"""The snap sub-package.

It contains Plonk the implementation of smoothed particle hydrodynamics
snapshot file.
"""

from .readers import load_snap
from .snap import Snap, SnapLike

__all__ = ['Snap', 'SnapLike', 'load_snap']



"""Reader for snapshot files."""

from pathlib import Path
from typing import Union

from ..snap import Snap
from .phantom_hdf5 import PhantomHDF5Snap

_available_formats = ('phantom',)


def load_snap(filename: Union[str, Path], filetype: str = 'phantom') -> Snap:
    """Load a snapshot from file.

    Parameters
    ----------
    filename
        Path to snapshot file.
    filetype
        The SPH file type.
    """
    if filetype.lower() in ('phantom', 'phantom_h5', 'phantom_hdf', 'phantom_hdf5'):
        filetype = 'phantom'
    if filetype not in _available_formats:
        raise ValueError('Unknown file type')
    if filetype == 'phantom':
        return PhantomHDF5Snap().generate_snap_from_file(filename)
    else:
        raise ValueError('Cannot load snapshot')



"""Radial profiles.

Heavily inspired by pynbody (https://pynbody.github.io/).
"""

from bisect import bisect
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame

from .. import Quantity
from ..snap import Snap
from ..utils.math import norm
from .particles import eccentricity as _eccentricity
from .particles import specific_angular_momentum


class Profile:
    """Radial profiles.

    A radial profile is a binning of particles in cylindrical or
    spherical shells around the origin, i.e. (0, 0, 0). For cylindrical
    profiles the cylindrical cells are perpendicular to the xy-plane,
    i.e. the bins are averaged azimuthally and in the z-direction.

    Parameters
    ----------
    snap
        The Snap object.
    ndim : optional
        The dimension of the profile. For ndim == 2, the radial binning
        is cylindrical in the xy-plane. For ndim == 3, the radial
        binning is spherical. Default is 2.
    radius_min : optional
        The minimum radius for binning. Defaults to minimum on the
        particles.
    radius_max : optional
        The maximum radius for binning. Defaults to the 99 percentile
        distance.
    n_bins : optional
        The number of radial bins. Default is 100.
    spacing : optional
        The spacing of radial bins. Can be 'linear' or 'log'. Default is
        'linear'.
    ignore_accreted : optional
        Ignore particles accreted onto sinks. Default is True.

    Examples
    --------
    Generate profile from snapshot.

    >>> prof = plonk.Profile(snap=snap)
    >>> prof = plonk.Profile(snap=snap, n_bins=300)
    >>> prof = plonk.Profile(snap=snap, radius_min=1, radius_max=300)
    >>> prof = plonk.Profile(snap=snap, spacing='log')

    If snap has physical units and setting 'radius_min' or 'radius_max'
    must use physical quantities.

    >>> radius_min = plonk.Quantity('1 au')
    >>> radius_max = plonk.Quantity('300 au')
    >>> prof = plonk.Profile(
    ...     snap=snap,
    ...     radius_min=radius_min,
    ...     radius_max=radius_max
    ... )

    To access a profile.

    >>> prof['density']
    >>> prof['scale_height']

    To set a new profile.

    >>> prof['aspect_ratio'] = prof['scale_height'] / prof['radius']

    Alternatively use the profile_property decorator.

    >>> @Profile.profile_property
    ... def mass(prof):
    ...     M = prof.snap['mass']
    ...     return prof.particles_to_binned_quantity(np.sum, M)

    Plot one or many quantities on the profile.

    >>> prof.plot('radius', 'density')
    >>> prof.plot('radius', ['angmom_x', angmom_y'])

    Plot a quantity on the profile with units.

    >>> prof.plot('radius', 'density', x_unit='au', y_unit='g/cm**2')
    """

    _profile_functions: Dict[str, Callable] = {}

    @staticmethod
    def profile_property(fn: Callable) -> Callable:
        """Decorate function to add profile to Profile.

        Parameters
        ----------
        fn
            A function that returns the profile as an array. The name of
            the function is the string with which to reference the array.

        Returns
        -------
        Callable
            The function which returns the array.
        """
        Profile._profile_functions[fn.__name__] = fn
        return fn

    def __init__(
        self,
        snap: Snap,
        ndim: Optional[int] = 2,
        radius_min: Optional[Any] = None,
        radius_max: Optional[Any] = None,
        n_bins: int = 100,
        spacing: str = 'linear',
        ignore_accreted: bool = True,
    ):

        self.snap = snap
        self.ndim = ndim
        self.spacing = self._check_spacing(spacing)
        self.properties: Dict[str, Any] = {}

        self._profiles: Dict[str, ndarray] = {}

        self._mask = self._setup_particle_mask(ignore_accreted)
        self._x = self._calculate_x()
        self.range = self._set_range(radius_min, radius_max)
        self.n_bins = n_bins

        self.bin_edges, self['size'] = self._setup_bins()
        self.bin_centers = 0.5 * (self.bin_edges[:-1] + self.bin_edges[1:])
        if self.snap._physical_units:
            self._particle_bin = np.digitize(
                self._x.magnitude, self.bin_edges.magnitude
            )
        else:
            self._particle_bin = np.digitize(self._x, self.bin_edges)
        self.bin_indicies = self._set_particle_bin_indicies()

        self._profiles['radius'] = self.bin_centers
        if self.snap._physical_units:
            self._profiles['number'] = np.histogram(
                self._x.magnitude, self.bin_edges.magnitude
            )[0]
        else:
            self._profiles['number'] = np.histogram(self._x, self.bin_edges)[0]

    def _check_spacing(self, spacing):
        if spacing.lower() in ('lin', 'linear'):
            spacing = 'linear'
        elif spacing.lower() in ('log', 'logarithm', 'logarithmic'):
            spacing = 'log'
        else:
            raise ValueError('Cannot determine spacing')
        return spacing

    def _setup_particle_mask(self, ignore_accreted: bool) -> ndarray:
        if ignore_accreted is False:
            return np.ones(len(self.snap), dtype=bool)
        h: ndarray = self.snap['h']
        return h > 0

    def _calculate_x(self) -> ndarray:
        pos: ndarray = self.snap['xyz']
        pos = pos[self._mask]
        if self.ndim == 2:
            return np.sqrt(pos[:, 0] ** 2 + pos[:, 1] ** 2)
        elif self.ndim == 3:
            return np.sqrt(pos[:, 0] ** 2 + pos[:, 1] ** 2 + pos[:, 2] ** 2)

    def _set_range(
        self, radius_min: Optional[Any], radius_max: Optional[Any]
    ) -> Tuple[float, float]:
        if self.snap._physical_units:
            return self._set_range_physical_units(radius_min, radius_max)
        else:
            return self._set_range_code_units(radius_min, radius_max)

    def _set_range_code_units(
        self, radius_min: Optional[float], radius_max: Optional[float]
    ) -> Tuple[float, float]:
        if radius_min is None:
            rmin = self._x.min()
        else:
            rmin = radius_min
        if radius_max is None:
            rmax = np.percentile(self._x, 99, axis=0)
        else:
            rmax = radius_max

        return rmin, rmax

    def _set_range_physical_units(
        self, radius_min: Optional[Any], radius_max: Optional[Any]
    ) -> Tuple[float, float]:
        if radius_min is None:
            rmin = self._x.min()
        else:
            if not isinstance(radius_min, Quantity):
                raise ValueError(
                    'Snap has physical units: must use dimensional radius_min'
                )
            rmin = radius_min.to_base_units()
        if radius_max is None:
            rmax = np.percentile(self._x.magnitude, 99, axis=0) * self._x.units
        else:
            if not isinstance(radius_max, Quantity):
                raise ValueError(
                    'Snap has physical units: must use dimensional radius_max'
                )
            rmax = radius_max.to_base_units()

        return rmin, rmax

    def _setup_bins(self) -> ndarray:
        if self.snap._physical_units:
            bin_edges = self._bin_edges_physical_units()
        else:
            bin_edges = self._bin_edges_code_units()
        if self.ndim == 2:
            bin_sizes = np.pi * (bin_edges[1:] ** 2 - bin_edges[:-1] ** 2)
        elif self.ndim == 3:
            bin_sizes = 4 / 3 * np.pi * (bin_edges[1:] ** 3 - bin_edges[:-1] ** 3)
        return bin_edges, bin_sizes

    def _bin_edges_physical_units(self):
        if self.spacing == 'linear':
            bin_edges = (
                np.linspace(
                    self.range[0].magnitude, self.range[1].magnitude, self.n_bins + 1
                )
                * self.range[0].units
            )
        elif self.spacing == 'log':
            bin_edges = (
                np.logspace(
                    np.log10(self.range[0].magnitude),
                    np.log10(self.range[1].magnitude),
                    self.n_bins + 1,
                )
                * self.range[0].units
            )
        else:
            raise ValueError('Cannot determine spacing to setup bins')
        return bin_edges

    def _bin_edges_code_units(self):
        if self.spacing == 'linear':
            bin_edges = np.linspace(self.range[0], self.range[1], self.n_bins + 1)
        elif self.spacing == 'log':
            bin_edges = np.logspace(
                np.log10(self.range[0]), np.log10(self.range[1]), self.n_bins + 1
            )
        else:
            raise ValueError('Cannot determine spacing to setup bins')
        return bin_edges

    def _set_particle_bin_indicies(self) -> List[ndarray]:
        sortind = self._particle_bin.argsort()
        sort_pind = self._particle_bin[sortind]
        binind = list()
        prev_index = bisect(sort_pind, 0)
        for i in range(self.n_bins):
            new_index = bisect(sort_pind, i + 1)
            binind.append(np.sort(sortind[prev_index:new_index]))
            prev_index = new_index
        return binind

    def _get_profile(self, name: str, args: Optional[Tuple[Any, ...]] = None):
        """Get a profile by name."""
        if name in self._profiles:
            return self._profiles[name]

        elif name in Profile._profile_functions:
            if args is not None:
                self._profiles[name] = Profile._profile_functions[name](self, *args)
            else:
                self._profiles[name] = Profile._profile_functions[name](self)
            return self._profiles[name]

        else:
            raise ValueError('Profile not available')

    def __getitem__(self, name: str) -> ndarray:
        """Return the profile of a given kind."""
        return self._get_profile(name)

    def __setitem__(self, name: str, item: ndarray):
        """Set the profile directly."""
        if not isinstance(item, (ndarray, Quantity)):
            raise ValueError('"item" must be ndarray or pint Quantity')
        if item.shape[0] != self.n_bins:
            raise ValueError('Length of array does not match number of bins')
        if name in self.loaded_keys():
            raise ValueError(
                'Attempting to overwrite existing profile. To do so, first delete the '
                'profile\nwith del prof["profile"], then try again.'
            )
        elif name in self.available_keys():
            raise ValueError(
                'Attempting to set profile already available. '
                'See prof.available_keys().'
            )
        self._profiles[name] = item

    def __delitem__(self, name):
        """Delete a profile from memory."""
        del self._profiles[name]

    def __len__(self):
        """Length as number of bins."""
        return self.n_bins

    def __repr__(self):
        """Object repr method."""
        return f'<plonk.Profile: {self.n_bins} bins>'

    def loaded_keys(self):
        """Return a listing of loaded profiles."""
        return tuple(sorted(self._profiles.keys()))

    def available_keys(self):
        """Return a listing of available profiles."""
        loaded = list(self.loaded_keys())
        available = list(self._profile_functions.keys())
        return tuple(sorted(set(loaded + available)))

    def plot(
        self,
        x: str,
        y: Union[str, List[str]],
        x_unit: Optional[str] = None,
        y_unit: Optional[Union[str, List[str]]] = None,
        ax: Any = None,
    ):
        """Plot profile.

        Parameters
        ----------
        x
            The x axis to plot as a string.
        y
            The y axis to plot. Can be string or multiple as a list of
            strings.
        x_unit : optional
            The x axis quantity unit as a string. Only works if using
            physical units.
        y_unit : optional
            The y axis quantity unit as a string or list of strings.
            Only works if using physical units.
        ax : optional
            A matplotlib Axes object to plot to.
        """
        if x not in self.available_keys():
            raise ValueError('Cannot determine x axis to plot')

        if isinstance(y, str):
            y = [y]

        if not self.snap._physical_units:
            if x_unit is not None or y_unit is not None:
                raise ValueError('Cannot set unit if snap is not in physical units')
        else:
            if y_unit is not None:
                if isinstance(y_unit, str):
                    y_unit = [y_unit]
                if len(y) != len(y_unit):
                    raise ValueError('Length of y does not match length of y_unit')

        _x = self[x]
        if x_unit is not None:
            _x = _x.to(x_unit)

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        xlabel = x.capitalize().replace('_', ' ')
        if self.snap._physical_units:
            xlabel = ' '.join([xlabel, f'[{_x.units:~P}]'])
        ax.set_xlabel(xlabel)

        for idx, yi in enumerate(y):
            if yi not in self.available_keys():
                raise ValueError(f'Cannot determine y axis to plot: {yi} not available')
            _y = self[yi]
            label = yi.capitalize().replace('_', ' ')
            if self.snap._physical_units:
                if y_unit is not None:
                    _y = _y.to(y_unit[idx])
                label = ' '.join([label, f'[{_y.units:~P}]'])
                ax.plot(_x.magnitude, _y.magnitude, label=label)
            else:
                ax.plot(_x, _y, label=label)

        ax.legend()

        return fig, ax

    def to_dataframe(self, all_available: bool = False) -> DataFrame:
        """Convert Profile to DataFrame.

        Parameters
        ----------
        all_available
            If True, this will calculate all available profiles before
            converting to a DataFrame.
        """
        if all_available:
            columns = self.available_keys()
        else:
            columns = self.loaded_keys()
        data = dict()
        for column in columns:
            data[column] = self[column]
        return pd.DataFrame(data)

    def particles_to_binned_quantity(self, function, *args) -> ndarray:
        """Calculate binned quantities from particles.

        This takes care of the bin indices and ignoring accreted
        particles (if requested in instantiating the profile).

        Parameters
        ----------
        function
            The function that acts on particles in the radial bin. The
            function should produce a single number, i.e. it is a
            reduction. E.g. to calculate the mean of a particle quantity
            in a radial bin use np.mean.
        *args
            The arguments to the function. Typically this is one or more
            particle arrays or other quantities on a snapshot.
        """
        binned_quantity = np.zeros(self.n_bins)
        for idx, bin_ind in enumerate(self.bin_indicies):
            val = function(*tuple([arg[self._mask][bin_ind] for arg in args]))
            if self.snap._physical_units:
                binned_quantity[idx] = val.magnitude
            else:
                binned_quantity[idx] = val
        if self.snap._physical_units:
            binned_quantity *= val.units
        return binned_quantity


@Profile.profile_property
def mass(prof) -> ndarray:
    """Mass profile."""
    M = prof.snap['mass']
    return prof.particles_to_binned_quantity(np.sum, M)


@Profile.profile_property
def density(prof) -> ndarray:
    """Density profile.

    Units are [mass / length ** ndim], which depends on ndim of profile.
    """
    return prof['mass'] / prof['size']


@Profile.profile_property
def smooth(prof) -> ndarray:
    """Smoothing length profile."""
    h = prof.snap['smooth']
    return prof.particles_to_binned_quantity(np.mean, h)


@Profile.profile_property
def scale_height(prof) -> ndarray:
    """Scale height profile."""
    z = prof.snap['z']
    return prof.particles_to_binned_quantity(np.std, z)


@Profile.profile_property
def aspect_ratio(prof) -> ndarray:
    """Aspect ratio profile."""
    H = prof['scale_height']
    R = prof['radius']
    return H / R


@Profile.profile_property
def angmom_mag(prof) -> ndarray:
    """Magnitude of specific angular momentum profile."""
    angmom = specific_angular_momentum(snap=prof.snap)
    angmom_mag = norm(angmom, axis=1)

    return prof.particles_to_binned_quantity(np.mean, angmom_mag)


@Profile.profile_property
def angmom_x(prof) -> ndarray:
    """x-component of specific angular momentum profile."""
    angmom = specific_angular_momentum(snap=prof.snap)
    angmom_x = angmom[:, 0]

    return prof.particles_to_binned_quantity(np.mean, angmom_x)


@Profile.profile_property
def angmom_y(prof) -> ndarray:
    """y-component of specific angular momentum profile."""
    angmom = specific_angular_momentum(snap=prof.snap)
    angmom_y = angmom[:, 1]

    return prof.particles_to_binned_quantity(np.mean, angmom_y)


@Profile.profile_property
def angmom_z(prof) -> ndarray:
    """z-component of specific angular momentum profile."""
    angmom = specific_angular_momentum(snap=prof.snap)
    angmom_z = angmom[:, 2]

    return prof.particles_to_binned_quantity(np.mean, angmom_z)


@Profile.profile_property
def angmom_theta(prof) -> ndarray:
    """Angle between specific angular momentum and xy-plane."""
    angmom_z = prof['angmom_z']
    angmom_mag = prof['angmom_mag']

    return np.arccos(angmom_z / angmom_mag)


@Profile.profile_property
def angmom_phi(prof) -> ndarray:
    """Angle between specific angular momentum and x-axis in xy-plane."""
    angmom_x = prof['angmom_x']
    angmom_y = prof['angmom_y']
    return np.arctan2(angmom_y, angmom_x)


@Profile.profile_property
def eccentricity(prof) -> ndarray:
    """Orbital eccentricity profile.

    This profile assumes the central mass is at (0, 0, 0) and that the
    gravitational parameter (mu = G M) is set.
    """
    try:
        gravitational_parameter = prof.properties['gravitational_parameter']
    except KeyError:
        raise ValueError(
            'To generate eccentricity profile, first set the gravitational parameter\n'
            'via prof.properties["gravitational_parameter"].'
        )
    ecc = _eccentricity(
        snap=prof.snap, gravitational_parameter=gravitational_parameter,
    )
    return prof.particles_to_binned_quantity(np.mean, ecc)



"""Snap, SubSnap, Sinks classes for snapshot files.

The Snap class contains all information related to a smoothed particle
hydrodynamics simulation snapshot file. The SubSnap class is for
accessing a subset of particles in a Snap.
"""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from scipy.spatial.transform import Rotation

from .. import Quantity


class _SinkUtility:
    def __init__(self, fn):
        self.fn = fn

    def __getitem__(self, inp):
        return self.fn(inp, sinks=True)

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.snap sinks>'


class Snap:
    """Smoothed particle hydrodynamics Snap object.

    Snapshot files contain the state of the simulation at a point in
    time. Typical minimum data from a smoothed particle hydrodynamics
    simulation include the particle positions and smoothing length, from
    which the density field can be reconstructed, as well as the
    particle type. In addition, the particle velocities are required to
    restart the simulation.

    Other data stored in the snapshot file include equation of state,
    dust, and magnetic field information, as well as numerical
    quantities related to time-stepping.

    Examples
    --------
    To access arrays on the particles.

    >>> snap['position']
    >>> snap['density']

    To access sink arrays.

    >>> snap.sinks['position']
    >>> snap.sinks['spin']

    To access a subset of particles as a SubSnap.

    >>> subsnap = snap[:100]
    >>> subsnap = snap[snap['x'] > 0]
    >>> subsnap = snap['gas']

    To set a new array.

    >>> snap['r'] = np.sqrt(snap['x'] ** 2 + snap['y'] ** 2)

    Alternatively, define a function.

    >>> @plonk.Snap.add_array()
    ... def radius(snap) -> ndarray:
    ...     radius = np.hypot(snap['x'], snap['y'])
    ...     return radius

    Possibly with units.

    >>> @plonk.Snap.add_array(unit='length')
    ... def radius(snap) -> ndarray:
    ...     radius = np.hypot(snap['x'], snap['y'])
    ...     return radius

    Or, use an existing one.

    >>> snap['R'] = plonk.analysis.particles.radial_distance(snap)

    Set physical units. Arrays are now Pint quantities.

    >>> snap.physical_units()
    """

    _array_registry: Dict[str, Callable] = {}
    _sink_registry: Dict[str, Callable] = {}

    _array_name_mapper = {
        'xyz': 'position',
        'pos': 'position',
        'vxyz': 'velocity',
        'vel': 'velocity',
        'h': 'smooth',
        'm': 'mass',
        'rho': 'density',
        'Bxyz': 'magfield',
        'spinxyz': 'spin',
    }

    _array_split_mapper = {
        'x': ('position', 0),
        'y': ('position', 1),
        'z': ('position', 2),
        'vx': ('velocity', 0),
        'vy': ('velocity', 1),
        'vz': ('velocity', 2),
        'velx': ('velocity', 0),
        'vely': ('velocity', 1),
        'velz': ('velocity', 2),
        'Bx': ('magfield', 0),
        'By': ('magfield', 1),
        'Bz': ('magfield', 2),
        'sx': ('spin', 0),
        'sy': ('spin', 1),
        'sz': ('spin', 2),
    }

    _array_units = {
        'alpha': 'dimensionless',
        'density': 'density',
        'divv': 'frequency',
        'dt': 'time',
        'dustfrac': 'dimensionless',
        'dust_type': 'dimensionless',
        'magfield': 'magnetic_field',
        'mass': 'mass',
        'position': 'length',
        'poten': 'energy',
        'smooth': 'length',
        'spin': 'angular_momentum',
        'tstop': 'time',
        'type': 'dimensionless',
        'velocity': 'velocity',
    }

    _particle_type = {
        'gas': 1,
        'dust': 2,
        'boundary': 3,
        'star': 4,
        'darkmatter': 5,
        'bulge': 6,
    }

    @staticmethod
    def add_array(unit: str = None) -> Callable:
        """Decorate function to add array to Snap.

        This function decorates a function that returns an array. The
        name of the function is the string with which to reference the
        array.

        Parameters
        ----------
        unit
            A string to represent the units of the array. E.g. 'length'
            for a 'radius' array.

        Returns
        -------
        Callable
            The decorator which returns the array.
        """

        def _add_array(fn):
            Snap._array_units[fn.__name__] = unit
            Snap._array_registry[fn.__name__] = fn
            return fn

        return _add_array

    @staticmethod
    def add_alias(name: str, alias: str) -> None:
        """Add alias to array.

        Parameters
        ----------
        name
            The name of the array.
        alias
            The alias to reference the array.
        """
        Snap._array_name_mapper[alias] = name

    def __init__(self):

        self.properties = {}
        self.units = {}
        self._arrays = {}
        self._sinks = {}
        self._file_pointer = None
        self._num_particles = -1
        self._num_sinks = -1
        self._families = {key: None for key in Snap._particle_type.keys()}
        self.rotation = None
        self.translation = None
        self._physical_units = False

    def close_file(self):
        """Close access to underlying file."""
        self._file_pointer.close()

    def loaded_arrays(self, sinks: bool = False):
        """Return a list of loaded arrays."""
        if sinks:
            return tuple(sorted(self._sinks.keys()))
        return tuple(sorted(self._arrays.keys()))

    def available_arrays(self, sinks: bool = False):
        """Return a list of available arrays."""
        if sinks:
            loaded = self.loaded_arrays(sinks)
            registered = tuple(sorted(self._sink_registry.keys()))
        else:
            loaded = self.loaded_arrays()
            registered = tuple(sorted(self._array_registry.keys()))
        return tuple(sorted(set(loaded + registered)))

    @property
    def sinks(self):
        """Sink particle arrays."""
        return _SinkUtility(self._get_array)

    @property
    def num_particles(self):
        """Return number of particles."""
        if self._num_particles == -1:
            self._num_particles = len(self._array_registry['type'](self))
        return self._num_particles

    @property
    def num_sinks(self):
        """Return number of sinks."""
        if self._num_sinks == -1:
            try:
                self._num_sinks = len(self._sink_registry['mass'](self))
            except KeyError:
                self._num_sinks = 0
        return self._num_sinks

    def add_unit(self, name: str, unit: Any, unit_str: str):
        """Define a unit on an array.

        Parameters
        ----------
        name
            The name of the array.
        unit
            The Pint units Quantity.
        unit_str
            The unit string. See units attribute for units.

        Examples
        --------
        New array 'arr' with dimension 'length' and units 'cm'.

        >>> snap.add_unit('arr', plonk.units('cm'), 'length')
        """
        if name in self._array_units:
            raise ValueError('Array unit already defined on Snap')
        if unit_str not in self.units:
            self.units[unit_str] = unit
        self._array_units[name] = unit_str

    def physical_units(self, unset: bool = False) -> Snap:
        """Set/unset physical units.

        Uses Pint.

        Returns
        -------
        Snap
        """
        if self._physical_units:
            raise ValueError(
                'Physical units already set: snap.physical_units(unset=True) to unset.'
            )
        if unset:
            for arr in self.loaded_arrays():
                del self._arrays[arr]
            for arr in self.loaded_arrays(sinks=True):
                del self._sinks[arr]
            self._physical_units = False
        else:
            for arr in self.loaded_arrays():
                self._arrays[arr] = self._arrays[arr] * self.get_array_unit(arr)
            for arr in self.loaded_arrays(sinks=True):
                self._sinks[arr] = self._sinks[arr] * self.get_array_unit(arr)
            self._physical_units = True

        return self

    def rotate(self, rotation: Rotation) -> Snap:
        """Rotate snapshot.

        Parameters
        ----------
        rotation
            The rotation as a scipy.spatial.transform.Rotation object.

        Returns
        -------
        Snap
            The rotated Snap. Note that the rotation operation is
            in-place.
        """
        for arr in self._rotation_required():
            if arr in self.loaded_arrays():
                self._arrays[arr] = rotation.apply(self._arrays[arr])
            if arr in self.loaded_arrays(sinks=True):
                self._sinks[arr] = rotation.apply(self._sinks[arr])

        self.rotation = rotation

        return self

    def translate(self, translation: ndarray) -> Snap:
        """Translate snapshot.

        I.e. shift the snapshot origin.

        Parameters
        ----------
        translation
            The translation as a (3,) ndarray like (x, y, z).

        Returns
        -------
        Snap
            The translated Snap. Note that the translation operation is
            in-place.
        """
        if len(translation) != 3:
            raise ValueError('translation must be like (x, y, z)')
        if 'position' in self.loaded_arrays():
            self._arrays['position'] += translation
        if 'position' in self.loaded_arrays(sinks=True):
            self._sinks['position'] += translation

        self.translation = translation

        return self

    def to_dataframe(self, columns: Union[Tuple[str, ...], List[str]]) -> DataFrame:
        """Convert Snap to DataFrame.

        Parameters
        ----------
        columns
            A list of columns to add to the data frame.

        Returns
        -------
        DataFrame
        """
        d = dict()
        cols = list(columns)
        for col in cols:
            arr = self[col]
            arr = cast(ndarray, arr)
            if arr.ndim == 2:
                for idx in range(arr.shape[1]):
                    d[f'{col}.{idx+1}'] = arr[:, idx]
            else:
                d[col] = arr
        return pd.DataFrame(d)

    def _rotation_required(self):
        return set([val[0] for val in self._array_split_mapper.values()])

    def _get_family_indices(self, name: str):
        """Get a family by name."""
        if name in self._families:
            if self._families[name] is None:
                self._families[name] = np.flatnonzero(
                    self['type'] == Snap._particle_type[name]
                )
            return self._families[name]
        else:
            raise ValueError('Family not available')

    def get_array_unit(self, arr: str) -> Any:
        """Get array code units.

        Parameters
        ----------
        arr
            The string representing the quantity.

        Returns
        -------
        unit
            The Pint unit quantity, or the float 1.0 if no unit found.
        """
        if arr in self._array_split_mapper:
            arr = self._array_split_mapper[arr][0]
        elif arr in self._array_name_mapper:
            arr = self._array_name_mapper[arr]
        try:
            unit = self.units[self._array_units[arr]]
        except KeyError:
            unit = 1.0
        return unit

    def _get_array_from_registry(self, name: str, sinks: bool = False):
        if sinks:
            array = Snap._sink_registry[name](self)
            array_dict = self._sinks
        else:
            array = Snap._array_registry[name](self)
            array_dict = self._arrays
        if self.rotation is not None and name in self._rotation_required():
            array = self.rotation.apply(array)
        if self.translation is not None and name == 'position':
            array += self.translation
        if self._physical_units and not isinstance(array, Quantity):
            unit = self.get_array_unit(name)
            array_dict[name] = unit * array
        else:
            array_dict[name] = array

    def _get_array(
        self, name: str, index: Optional[int] = None, sinks: bool = False
    ) -> ndarray:
        """Get an array by name."""
        if name in self.available_arrays(sinks):
            name = name
        elif name in self._array_name_mapper.keys():
            name = self._array_name_mapper[name]
        elif name in self._array_split_mapper.keys():
            name, index = self._array_split_mapper[name]
        else:
            raise ValueError('Array not available')

        if sinks:
            array_dict = self._sinks
        else:
            array_dict = self._arrays
        if name in array_dict:
            if index is None:
                return array_dict[name]
            return array_dict[name][:, index]
        elif name in Snap._array_registry or name in Snap._sink_registry:
            self._get_array_from_registry(name, sinks)
            if index is None:
                return array_dict[name]
            return array_dict[name][:, index]
        else:
            raise ValueError('Array not available')

    def __getitem__(
        self, inp: Union[str, ndarray, int, slice]
    ) -> Union[ndarray, SubSnap]:
        """Return an array, or family, or subset."""
        if isinstance(inp, str):
            if inp in self._families:
                return SubSnap(self, self._get_family_indices(inp))
            elif inp in self.available_arrays():
                return self._get_array(inp)
            elif inp in self._array_name_mapper.keys():
                return self._get_array(inp)
            elif inp in self._array_split_mapper.keys():
                return self._get_array(inp)
        elif isinstance(inp, ndarray):
            if np.issubdtype(np.bool, inp.dtype):
                return SubSnap(self, np.flatnonzero(inp))
            elif np.issubdtype(np.int, inp.dtype):
                return SubSnap(self, inp)
        elif isinstance(inp, int):
            raise NotImplementedError
        elif isinstance(inp, slice):
            i1, i2, step = inp.start, inp.stop, inp.step
            if step is not None:
                return SubSnap(self, np.arange(i1, i2, step))
            return SubSnap(self, np.arange(i1, i2))
        raise ValueError('Cannot determine item to return')

    def __setitem__(self, name: str, item: ndarray):
        """Set an array."""
        if not isinstance(item, ndarray):
            raise ValueError('"item" must be ndarray')
        if item.shape[0] != len(self):
            raise ValueError('Length of array does not match particle number')
        if name in self.loaded_arrays():
            raise ValueError(
                'Attempting to overwrite existing array. To do so, first delete the '
                'array\nwith del snap["array"], then try again.'
            )
        elif (
            name in self.available_arrays()
            or name in self._array_split_mapper.keys()
            or name in self._array_name_mapper.keys()
        ):
            raise ValueError(
                'Attempting to set array already available. '
                'See snap.available_arrays().'
            )
        self._arrays[name] = item

    def __delitem__(self, name):
        """Delete an array from memory."""
        del self._arrays[name]

    def __len__(self):
        """Length as number of particles."""
        return self.num_particles

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.Snap>'


class SubSnap(Snap):
    """A Snap subset of particles.

    The sub-snap is generated via an index array.

    Parameters
    ----------
    base
        The base snapshot.
    indices
        A (N,) array of particle indices to include in the sub-snap.
    """

    def __init__(self, base: Snap, indices: ndarray):
        super().__init__()
        self.base = base
        self.properties = self.base.properties
        self._file_pointer = self.base._file_pointer
        self._indices = indices
        self._num_particles = len(indices)
        self._arrays = self.base._arrays

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.SubSnap>'

    def _get_array(
        self, name: str, index: Optional[int] = None, sinks: bool = False
    ) -> ndarray:
        return self.base._get_array(name, index, sinks)[self._indices]


SnapLike = Union[Snap, SubSnap]


def get_array_from_input(
    snap: SnapLike, inp: Union[str, ndarray], default: str = None
) -> ndarray:
    """Get array on Snap.

    Parameters
    ----------
    snap
        The Snap or SubSnap.
    inp
        The input as a string or ndarray. If a string return
        snap[inp], otherwise return inp as a ndarray.
    default
        The default array as a string resolved as snap[default].

    Returns
    -------
    ndarray
        The array on the particles.
    """
    if isinstance(inp, ndarray):
        return inp
    elif isinstance(inp, str):
        return get_array_in_code_units(snap, inp)
    elif default is not None:
        return get_array_in_code_units(snap, default)
    raise ValueError('Cannot determine array to return')


def get_array_in_code_units(snap: SnapLike, name: str) -> ndarray:
    """Get array in code units.

    Parameters
    ----------
    snap
        The Snap or SubSnap.
    name
        The array name.

    Returns
    -------
    ndarray
        The array on the particles in code units.
    """
    arr = snap[name]
    if isinstance(arr, Quantity):
        return (arr / snap.get_array_unit(name)).magnitude
    return arr



"""Geometry, and coordinate transformations."""

from typing import Tuple

import numpy as np
from numpy import ndarray
from scipy.interpolate import RectBivariateSpline
from skimage import transform


def cartesian_to_polar(
    interpolated_data_cartesian: ndarray,
    extent_cartesian: Tuple[float, float, float, float],
) -> Tuple[ndarray, Tuple[float, float, float, float]]:
    """Convert interpolated Cartesian pixel grid to polar coordinates.

    Parameters
    ----------
    interpolated_data_cartesian
        The interpolated data on a Cartesian grid.
    extent_cartesian
        The extent in Cartesian space as (xmin, xmax, ymin, ymax). It
        must be square.

    Returns
    -------
    interpolated_data_polar
        The interpolated data on a polar grid (R, phi).
    extent_polar
        The extent on a polar grid (0, Rmax, 0, 2π).
    """
    data, extent = interpolated_data_cartesian, extent_cartesian

    if not np.allclose(extent[1] - extent[0], extent[3] - extent[2]):
        raise ValueError('Bad polar plot: x and y have different scales')

    number_of_pixels = data.shape
    radius_pix = 0.5 * data.shape[0]

    data = transform.warp_polar(data, radius=radius_pix)

    radius = 0.5 * (extent[1] - extent[0])
    extent_polar = (0, radius, 0, 2 * np.pi)

    x_grid = np.linspace(*extent[:2], data.shape[0])
    y_grid = np.linspace(*extent[2:], data.shape[1])
    spl = RectBivariateSpline(x_grid, y_grid, data)
    x_regrid = np.linspace(extent[0], extent[1], number_of_pixels[0])
    y_regrid = np.linspace(extent[2], extent[3], number_of_pixels[1])
    interpolated_data_polar = spl(x_regrid, y_regrid)

    return interpolated_data_polar, extent_polar



"""Plot functions for visualization."""

from typing import Any

import matplotlib as mpl
import numpy as np
from numpy import ndarray

from ..snap import SnapLike
from .interpolation import Extent


def particle_plot(
    *, snap: SnapLike, x: ndarray, y: ndarray, extent: Extent, ax: Any, **kwargs,
):
    """Plot particles.

    Parameters
    ----------
    snap
        The Snap object to visualize.
    x
        The x-coordinates for the particle plot.
    y
        The y-coordinates for the particle plot.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.plot method.

    Returns
    -------
    lines
        A list of matplotlib Line2D objects.
    """
    h: ndarray = snap['smooth']
    mask = (
        (h > 0) & (x > extent[0]) & (x < extent[1]) & (y > extent[2]) & (y < extent[3])
    )
    fmt = kwargs.get('fmt', 'k.')
    lines = ax.plot(x[mask], y[mask], fmt, **kwargs)
    return lines


def render_plot(
    *, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs,
):
    """Plot 1d interpolated data as a rendered image.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    image
        A matplotlib AxesImage object.
    """
    try:
        norm = kwargs.pop('norm')
    except KeyError:
        norm = 'linear'
    if norm.lower() in ('linear', 'lin'):
        norm = mpl.colors.Normalize()
    elif norm.lower() in ('logarithic', 'logarithm', 'log', 'log10'):
        norm = mpl.colors.LogNorm()
    else:
        raise ValueError('Cannot determine normalization for colorbar')

    image = ax.imshow(
        interpolated_data, origin='lower', extent=extent, norm=norm, **kwargs
    )

    return image


def contour_plot(
    *, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs,
):
    """Plot 1d interpolated data as a contour plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    contour
        A matplotlib QuadContourSet object.
    """
    n_interp_x, n_interp_y = interpolated_data.shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y),
    )

    contour = ax.contour(X, Y, interpolated_data, **kwargs)

    return contour


def quiver_plot(
    *, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs,
):
    """Plot 2d interpolated data as a quiver plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    quiver
        A matplotlib Quiver object.
    """
    n_interp_x, n_interp_y = interpolated_data[0].shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y)
    )
    U, V = interpolated_data[0], interpolated_data[1]

    number_of_arrows = kwargs.pop('number_of_arrows', (25, 25))
    normalize_vectors = kwargs.pop('normalize_vectors', False)

    n_x, n_y = number_of_arrows[0], number_of_arrows[1]
    stride_x = int(n_interp_x / n_x)
    stride_y = int(n_interp_y / n_y)
    X = X[::stride_y, ::stride_x]
    Y = Y[::stride_y, ::stride_x]
    U = U[::stride_y, ::stride_x]
    V = V[::stride_y, ::stride_x]
    if normalize_vectors:
        norm = np.hypot(U, V)
        U /= norm
        V /= norm

    quiver = ax.quiver(X, Y, U, V, **kwargs)

    return quiver


def stream_plot(
    *, interpolated_data: ndarray, extent: Extent, ax: Any, **kwargs,
):
    """Plot 2d interpolated data as a stream plot.

    Parameters
    ----------
    interpolated_data
        The data interpolated to a pixel grid.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax).
    ax
        A matplotlib Axes handle.
    **kwargs
        Keyword arguments to pass to ax.imshow method.

    Returns
    -------
    streamplot
        A matplotlib StreamplotSet object.
    """
    n_interp_x, n_interp_y = interpolated_data[0].shape
    X, Y = np.meshgrid(
        np.linspace(*extent[:2], n_interp_x), np.linspace(*extent[2:], n_interp_y)
    )
    U, V = interpolated_data[0], interpolated_data[1]

    streamplot = ax.streamplot(X, Y, U, V, **kwargs)

    return streamplot



"""The analysis sub-package.

It contains Plonk implementations of typical smoothed particle
hydrodynamics post-simulation analysis tasks.

Examples
--------
Create a radial profile in the xy-plane.

>>> p = Profile(snap, radius_min=10, radius_max=200, n_bins=100)
>>> p.plot('radius', 'density')

Calculate the angular momentum on the particles.

>>> angmom = particles.angular_momentum(snap)

Calculate the total angular momentum over all particles.

>>> angmom_tot = total.angular_momentum(snap)

Calculate the Roche sphere radius given two sink particles.

>>> m1 = snap.sinks['mass'][0]
>>> m2 = snap.sinks['mass'][1]
>>> separation = np.linalg.norm(
        snap.sinks['position'][0] - snap.sinks['position'][1]
    )
>>> Roche = sinks.Roche_sphere(m1, m2, separation)
"""

from . import particles, sinks, total
from .profile import Profile

__all__ = ['Profile', 'particles', 'sinks', 'total']



"""Physical units on snapshots."""


def generate_units_dictionary(length, mass, time):
    """Generate units dictionary.

    Parameters
    ----------
    length
        Length unit as a Pint quantity.
    mass
        Mass unit as a Pint quantity.
    time
        Time unit as a Pint quantity.

    Returns
    -------
    units
        A dictionary of units as Pint quantities.
    """
    units = {}

    units['dimensionless'] = length / length
    units['length'] = length
    units['time'] = time
    units['mass'] = mass
    units['frequency'] = 1 / time
    units['velocity'] = length / time
    units['momentum'] = mass * length / time
    units['angular_momentum'] = mass * length ** 2 / time
    units['density'] = mass / length ** 3
    units['acceleration'] = length / time ** 2
    units['force'] = mass * length / time ** 2
    units['energy'] = mass * length ** 2 / time ** 2

    return units



"""Utility functions."""

from typing import Any, Union

import numpy as np

from .. import units


def is_documented_by(original):
    """Wrap function to add docstring."""

    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def time_string(
    snap,
    unit: Union[str, Any] = units('year'),
    unit_str: str = 'yr',
    float_format: str = '.0f',
):
    """Generate time stamp string.

    Parameters
    ----------
    snap
        The Snap object.
    unit
        The time unit. Can be a string to pass to Pint or a Pint unit.
        Default is plonk.units('year').
    unit_str
        The unit string to print. Default is 'yr'.
    float_format
        The format for the time float value. Default is '.0f'.

    Examples
    --------
    Generate a list of strings of snapshot time.

    >>> text = [time_string(snap) for snap in snaps]
    """
    time = snap.properties['time'] * snap.units['time'].to(unit).magnitude
    return f't = {time:{float_format}} {unit_str}'


def get_extent_from_percentile(snap, percentile=99):
    """Get extent from percentile.

    Calculate a xy-plane square box such that some percentile of
    particles is contained within a sphere inscribed in the box.

    Parameters
    ----------
    snap
        The Snap object.
    percentile : optional
        The percentile used in the calculation. Default is 99.

    Returns
    -------
    tuple
        The extent of the box as (xmin, xmax, ymin, ymax).
    """
    r = np.sqrt(snap['x'] ** 2 + snap['y'] ** 2 + snap['z'] ** 2)
    size = np.percentile(r, percentile)
    return (-size, size, -size, size)



"""Global (total) quantities.

Calculate various global (total) quantities on the particles.
"""

from typing import Tuple, Union

import numpy as np
from numpy import ndarray

from .. import Quantity
from ..snap import SnapLike
from ..utils.math import norm
from . import particles


def center_of_mass(snap: SnapLike, ignore_accreted: bool = True) -> ndarray:
    """Calculate the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    ndarray
        The center of mass as a vector (cx, cy, cz).
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        pos: ndarray = snap['position'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']

    return (mass[:, np.newaxis] * pos).sum(axis=0) / mass.sum()


def mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total mass.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
    else:
        mass = snap['mass']

    return mass.sum()


def gas_mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total gas mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total gas mass.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        dustfrac: ndarray = snap['dustfrac'][h > 0]
    else:
        mass = snap['mass']
        dustfrac = snap['dustfrac']

    gas_frac = 1 - dustfrac.sum(axis=1)
    return (mass * gas_frac).sum()


def dust_mass(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total dust mass per species.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total dust mass per species.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        dustfrac: ndarray = snap['dustfrac'][h > 0]
    else:
        mass = snap['mass']
        dustfrac = snap['dustfrac']

    return (mass[:, np.newaxis] * dustfrac).sum(axis=0)


def accreted_mass(snap: SnapLike) -> float:
    """Calculate the accreted mass.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    float
        The accreted mass.
    """
    h: ndarray = snap['smooth']
    mass: ndarray = snap['mass'][~(h > 0)]

    return mass.sum()


def momentum(snap: SnapLike, ignore_accreted: bool = True) -> ndarray:
    """Calculate the total momentum.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    ndarray
        The total linear momentum like (px, py, pz).
    """
    return particles.momentum(snap=snap, ignore_accreted=ignore_accreted).sum(axis=0)


def angular_momentum(
    snap: SnapLike,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = True,
) -> ndarray:
    """Calculate the total angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    ndarray
        The total angular momentum like (lx, ly, lz).
    """
    return particles.angular_momentum(
        snap=snap, origin=origin, ignore_accreted=ignore_accreted
    ).sum(axis=0)


def specific_angular_momentum(
    snap: SnapLike,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = True,
) -> ndarray:
    """Calculate the total specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    ndarray
        The total specific angular momentum on the particles like
        (hx, hy, hz).
    """
    return particles.specific_angular_momentum(
        snap=snap, origin=origin, ignore_accreted=ignore_accreted
    ).sum(axis=0)


def kinetic_energy(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total kinetic energy.
    """
    return particles.kinetic_energy(snap=snap, ignore_accreted=ignore_accreted).sum(
        axis=0
    )


def specific_kinetic_energy(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the total specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The total specific kinetic energy.
    """
    return particles.specific_kinetic_energy(
        snap=snap, ignore_accreted=ignore_accreted
    ).sum(axis=0)


def inclination(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the inclination with respect to the xy-plane.

    The inclination is calculated by taking the angle between the
    angular momentum vector and the z-axis, with the angular momentum
    calculated with respect to the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The mean inclination.
    """
    angmom = angular_momentum(snap=snap, ignore_accreted=ignore_accreted)
    return np.arccos(angmom[2] / norm(angmom))


def position_angle(snap: SnapLike, ignore_accreted: bool = True) -> float:
    """Calculate the position angle of inclination.

    The position angle is taken from the x-axis in the xy-plane. It
    defines a unit vector around which the snap is inclined.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is True.

    Returns
    -------
    float
        The mean inclination.
    """
    angmom = angular_momentum(snap=snap, ignore_accreted=ignore_accreted)
    if isinstance(angmom, Quantity):
        pi_2 = np.pi / 2 * Quantity('radian')
    else:
        pi_2 = np.pi / 2
    return np.arctan2(angmom[1], angmom[0]) + pi_2



"""Splash interpolation functions.

Derived from Splash: https://github.com/danieljprice/splash.
"""

import numba
import numpy as np
from numpy import ndarray

RADKERNEL = 2.0
RADKERNEL2 = 4.0
CNORMK3D = 1.0 / np.pi

NPTS = 100
MAXCOLTABLE = 1000
DQ2TABLE = RADKERNEL2 / MAXCOLTABLE
DDQ2TABLE = 1.0 / DQ2TABLE

IVERBOSE = -1


@numba.njit
def w_cubic(q2: float):
    """Cubic spline kernel.

    Parameters
    ----------
    q2

    Returns
    -------
    w
    """
    w = 0.0
    if q2 < 1.0:
        q = np.sqrt(q2)
        w = 1.0 - 1.5 * q2 + 0.75 * q2 * q
    elif q2 < 4:
        q = np.sqrt(q2)
        w = 0.25 * (2.0 - q) ** 3
    return w


@numba.njit
def setup_integratedkernel():
    """Set up integrated kernel.

    Tabulates the integral through the cubic spline kernel tabulated in
    (r/h)**2 so that sqrt is not necessary.

    Returns
    -------
    coltable
    """
    coltable = np.zeros(MAXCOLTABLE + 1)

    for idx in range(MAXCOLTABLE):
        # Tabulate for (cylindrical) r**2 between 0 and RADKERNEL**2
        rxy2 = idx * DQ2TABLE

        # Integrate z between 0 and sqrt(RADKERNEL^2 - rxy^2)
        deltaz = np.sqrt(RADKERNEL2 - rxy2)
        dz = deltaz / (NPTS - 1)
        coldens = 0
        for j in range(NPTS):
            z = j * dz
            q2 = rxy2 + z * z
            wkern = w_cubic(q2)
            if j == 0 or j == NPTS - 1:
                coldens = coldens + 0.5 * wkern * dz
            else:
                coldens = coldens + wkern * dz
        coltable[idx] = 2.0 * coldens * CNORMK3D
    coltable[MAXCOLTABLE] = 0.0

    return coltable


@numba.njit
def wfromtable(q2, coltable):
    """Interpolate from integrated kernel table values to give w(q).

    Parameters
    ----------
    q2
    coltable

    Returns
    -------
    w
    """
    # Find nearest index in table
    index = int(q2 * DDQ2TABLE)
    index1 = min(index, MAXCOLTABLE)

    # Find increment along from this index
    dxx = q2 - index * DQ2TABLE

    # Find gradient
    dwdx = (coltable[index1] - coltable[index]) * DDQ2TABLE

    # Compute value of integrated kernel
    wfromtable = coltable[index] + dwdx * dxx

    return wfromtable


@numba.njit
def interpolate_projection(
    x: ndarray,
    y: ndarray,
    hh: ndarray,
    weight: ndarray,
    dat: ndarray,
    itype: ndarray,
    npart: int,
    xmin: float,
    ymin: float,
    npixx: int,
    npixy: int,
    pixwidthx: float,
    pixwidthy: float,
    normalise: bool,
):
    """Interpolate particles to grid via projection.

    Parameters
    ----------
    x
        The particle x positions.
    y
        The particle y positions.
    hh
        The particle smoothing length.
    weight
        The particle weight.
    dat
        The scalar data to interpolate.
    itype
        The particle type.
    npart
        The number of particles.
    xmin
        The minimum x position.
    ymin
        The minimum y position.
    npixx
        The number of pixels in the x direction.
    npixy
        The number of pixels in the y direction.
    normalise
        Whether to normalize.

    Return
    ------
    datsmooth
        The data smoothed to a pixel grid.
    """
    coltable = setup_integratedkernel()

    datsmooth = np.zeros((npixx, npixy))
    datnorm = np.zeros((npixx, npixy))
    dx2i = np.zeros(npixx)
    xpix = np.zeros(npixx)
    term = 0.0

    xminpix = xmin - 0.5 * pixwidthx
    yminpix = ymin - 0.5 * pixwidthy
    xmax = xmin + npixx * pixwidthx
    ymax = ymin + npixy * pixwidthy

    # Use a minimum smoothing length on the grid to make sure that particles
    # contribute to at least one pixel
    hmin = 0.5 * max(pixwidthx, pixwidthy)

    xpix = xminpix + np.arange(1, npixx + 1) * pixwidthx
    nsubgrid = 0
    nok = 0
    hminall = 1e10

    # Loop over particles
    for idx in range(npart):

        # Skip particles with itype < 0
        if itype[idx] < 0:
            continue

        # Set h related quantities
        hi = hh[idx]
        horigi = hi
        if not hi > 0.0:
            continue

        # Radius of the smoothing kernel
        radkern = RADKERNEL * hi

        # Cycle as soon as we know the particle does not contribute
        xi = x[idx]
        xpixmin = xi - radkern
        if xpixmin > xmax:
            continue
        xpixmax = xi + radkern
        if xpixmax < xmin:
            continue

        yi = y[idx]
        ypixmin = yi - radkern
        if ypixmin > ymax:
            continue
        ypixmax = yi + radkern
        if ypixmax < ymin:
            continue

        # Take resolution length as max of h and 1/2 pixel width
        if hi < hmin:
            hminall = min(hi, hminall)
            nsubgrid = nsubgrid + 1
            hsmooth = hmin
        else:
            hsmooth = hi
            nok = nok + 1
        radkern = RADKERNEL * hsmooth

        # Set kernel related quantities
        hi1 = 1.0 / hsmooth
        hi21 = hi1 * hi1
        termnorm = weight[idx] * horigi
        term = termnorm * dat[idx]

        # Loop over pixels, adding the contribution from this particle
        # copy by quarters if all pixels within domain
        ipixmin = int((xi - radkern - xmin) / pixwidthx)
        ipixmax = int((xi + radkern - xmin) / pixwidthx) + 1
        jpixmin = int((yi - radkern - ymin) / pixwidthy)
        jpixmax = int((yi + radkern - ymin) / pixwidthy) + 1

        # Make sure they only contribute to pixels in the image
        # (note that this optimises much better than using min/max)
        if ipixmin < 0:
            ipixmin = 0
        if jpixmin < 0:
            jpixmin = 0
        if ipixmax > npixx:
            ipixmax = npixx
        if jpixmax > npixy:
            jpixmax = npixy

        # Precalculate an array of dx2 for this particle (optimisation)
        for ipix in range(ipixmin, ipixmax):
            dx2i[ipix] = ((xpix[ipix] - xi) ** 2) * hi21

        for jpix in range(jpixmin, jpixmax):
            ypix = yminpix + jpix * pixwidthy
            dy = ypix - yi
            dy2 = dy * dy * hi21
            for ipix in range(ipixmin, ipixmax):
                # dx2 pre-calculated; dy2 pre-multiplied by hi21
                q2 = dx2i[ipix] + dy2
                # SPH kernel - integral through cubic spline
                # interpolate from a pre-calculated table
                if q2 < RADKERNEL2:
                    wab = wfromtable(q2, coltable)
                    # Calculate data value at this pixel using the summation interpolant
                    datsmooth[ipix, jpix] = datsmooth[ipix, jpix] + term * wab
                    if normalise:
                        datnorm[ipix, jpix] = datnorm[ipix, jpix] + termnorm * wab

    # Normalise dat array
    if normalise:
        # Normalise everywhere (required if not using SPH weighting)
        for idxi in range(npixx):
            for idxj in range(npixy):
                if datnorm[idxi, idxj] > 0.0:
                    datsmooth[idxi, idxj] /= datnorm[idxi, idxj]

    # Warn about subgrid interpolation
    if nsubgrid > 1:
        nfull = int((xmax - xmin) / (hminall)) + 1
        if nsubgrid > 0.1 * nok and IVERBOSE > -1:
            print('Warning: pixel size > 2h for ', nsubgrid, ' particles')
            print('need ', nfull, ' pixels for full resolution')

    # Return datsmooth
    return datsmooth.T


@numba.njit
def interpolate_cross_section(
    x: ndarray,
    y: ndarray,
    z: ndarray,
    hh: ndarray,
    weight: ndarray,
    dat: ndarray,
    itype: ndarray,
    npart: int,
    xmin: float,
    ymin: float,
    zslice: float,
    npixx: int,
    npixy: int,
    pixwidthx: float,
    pixwidthy: float,
    normalise: bool,
):
    """Interpolate particles to grid via cross section.

    Parameters
    ----------
    x
        The particle x positions.
    y
        The particle y positions.
    z
        The particle z positions.
    hh
        The particle smoothing length.
    weight
        The particle weight.
    dat
        The scalar data to interpolate.
    itype
        The particle type.
    npart
        The number of particles.
    xmin
        The minimum x position.
    ymin
        The minimum y position.
    zslice
        Cross section location.
    npixx
        The number of pixels in the x direction.
    npixy
        The number of pixels in the y direction.
    normalise
        Whether to normalize.

    Return
    ------
    datsmooth
        The data smoothed to a pixel grid.
    """
    datsmooth = np.zeros((npixx, npixy))
    datnorm = np.zeros((npixx, npixy))
    dx2i = np.zeros(npixx)
    const = CNORMK3D

    # Loop over particles
    for idx in range(npart):

        # Skip particles with itype < 0
        if itype[idx] < 0:
            continue

        # Set h related quantities
        hi = hh[idx]
        if not hi > 0.0:
            continue
        hi1 = 1.0 / hi
        hi21 = hi1 * hi1
        radkern = RADKERNEL * hi

        # For each particle, work out distance from the cross section slice
        dz = zslice - z[idx]
        dz2 = dz ** 2 * hi21

        # If this is < 2h then add the particle's contribution to the pixels
        # otherwise skip all this and start on the next particle
        if dz2 < RADKERNEL2:

            xi = x[idx]
            yi = y[idx]
            termnorm = const * weight[idx]
            term = termnorm * dat[idx]

            # Loop over pixels, adding the contribution from this particle
            # copy by quarters if all pixels within domain
            ipixmin = int((xi - radkern - xmin) / pixwidthx)
            ipixmax = int((xi + radkern - xmin) / pixwidthx) + 1
            jpixmin = int((yi - radkern - ymin) / pixwidthy)
            jpixmax = int((yi + radkern - ymin) / pixwidthy) + 1

            # Make sure they only contribute to pixels in the image
            # (note that this optimises much better than using min/max)
            if ipixmin < 0:
                ipixmin = 0
            if jpixmin < 0:
                jpixmin = 0
            if ipixmax > npixx:
                ipixmax = npixx
            if jpixmax > npixy:
                jpixmax = npixy

            # Precalculate an array of dx2 for this particle (optimisation)
            for ipix in range(ipixmin, ipixmax):
                dx2i[ipix] = ((xmin + (ipix - 0.5) * pixwidthx - xi) ** 2) * hi21 + dz2

            # Loop over pixels, adding the contribution from this particle
            for jpix in range(jpixmin, jpixmax):
                ypix = ymin + (jpix - 0.5) * pixwidthy
                dy = ypix - yi
                dy2 = dy * dy * hi21
                for ipix in range(ipixmin, ipixmax):
                    q2 = dx2i[ipix] + dy2
                    # SPH kernel - cubic spline
                    if q2 < RADKERNEL2:
                        wab = w_cubic(q2)
                        # Calculate data value at this pixel using the summation
                        # interpolant
                        datsmooth[ipix, jpix] = datsmooth[ipix, jpix] + term * wab
                        if normalise:
                            datnorm[ipix, jpix] = datnorm[ipix, jpix] + termnorm * wab

    # Normalise dat array
    if normalise:
        # Normalise everywhere (required if not using SPH weighting)
        for idxi in range(npixx):
            for idxj in range(npixy):
                if datnorm[idxi, idxj] > 0.0:
                    datsmooth[idxi, idxj] /= datnorm[idxi, idxj]

    # Return datsmooth
    return datsmooth.T



"""Particle quantities.

Calculate various quantities on the particles.
"""

from typing import Any, Tuple, Union

import numpy as np
from numpy import ndarray

from .. import Quantity
from ..snap import SnapLike
from ..utils.math import cross, norm


def momentum(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the momentum.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The linear momentum on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    return mass[:, np.newaxis] * vel


def angular_momentum(
    snap: SnapLike,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = False,
) -> ndarray:
    """Calculate the angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The angular momentum on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']
        vel = snap['velocity']

    origin = np.array(origin)
    pos = pos - origin

    return mass[:, np.newaxis] * cross(pos, vel)


def specific_angular_momentum(
    snap: SnapLike,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = False,
) -> ndarray:
    """Calculate the specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The specific angular momentum on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = np.array(origin)
    pos = pos - origin

    return cross(pos, vel)


def kinetic_energy(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The kinetic energy on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    return 1 / 2 * mass * norm(vel, axis=1) ** 2


def specific_kinetic_energy(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The specific kinetic energy on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        vel: ndarray = snap['velocity'][h > 0]
    else:
        vel = snap['velocity']

    return 1 / 2 * norm(vel, axis=1) ** 2


def semi_major_axis(
    snap: SnapLike,
    gravitational_parameter: Any,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = False,
) -> ndarray:
    """Calculate the semi-major axis.

    The semi-major axis of particles around a mass specified by
    gravitational parameter with an optional to specify the position of
    the mass.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M). Can be a float or a Pint
        quantity.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The semi-major axis on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = np.array(origin)
    pos = pos - origin

    if isinstance(pos, Quantity):
        mu = gravitational_parameter.to_reduced_units()
    else:
        mu = gravitational_parameter

    radius = norm(pos, axis=1)

    specific_angular_momentum = cross(pos, vel)
    specific_angular_momentum_magnitude = norm(specific_angular_momentum, axis=1)

    specific_kinetic_energy = 1 / 2 * norm(vel, axis=1) ** 2
    specific_potential_energy = -mu / radius
    specific_energy = specific_kinetic_energy + specific_potential_energy

    term = specific_energy * (specific_angular_momentum_magnitude / mu) ** 2

    eccentricity = np.sqrt(1 + 2 * term.magnitude)

    return specific_angular_momentum_magnitude ** 2 / (mu * (1 - eccentricity ** 2))


def eccentricity(
    snap: SnapLike,
    gravitational_parameter: Any,
    origin: Union[ndarray, Tuple[float, float, float]] = (0.0, 0.0, 0.0),
    ignore_accreted: bool = False,
) -> ndarray:
    """Calculate the eccentricity.

    The eccentricity of particles around a mass specified by
    gravitational parameter with an optional to specify the position of
    the mass.

    Parameters
    ----------
    snap
        The Snap object.
    gravitational_parameter
        The gravitational parameter (mu = G M). Can be a float or a Pint
        quantity.
    origin : optional
        The origin around which to compute the angular momentum as a
        ndarray or tuple (x, y, z). Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The eccentricity on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    origin = np.array(origin)
    pos = pos - origin

    if isinstance(pos, Quantity):
        mu = gravitational_parameter.to_reduced_units()
    else:
        mu = gravitational_parameter

    radius = norm(pos, axis=1)

    specific_angular_momentum = cross(pos, vel)
    specific_angular_momentum_magnitude = norm(specific_angular_momentum, axis=1)

    specific_kinetic_energy = 1 / 2 * norm(vel, axis=1) ** 2
    specific_potential_energy = -mu / radius
    specific_energy = specific_kinetic_energy + specific_potential_energy

    term = specific_energy * (specific_angular_momentum_magnitude / mu) ** 2
    eccentricity = np.sqrt(1 + 2 * term)

    if isinstance(pos, Quantity):
        return eccentricity.magnitude
    return eccentricity


def inclination(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the inclination with respect to the xy-plane.

    The inclination is calculated by taking the angle between the
    angular momentum vector and the z-axis, with the angular momentum
    calculated with respect to the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The inclination on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        mass: ndarray = snap['mass'][h > 0]
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']
        vel = snap['velocity']

    origin = (mass[:, np.newaxis] * pos).sum(axis=0) / mass.sum()
    pos = pos - origin

    specific_angular_momentum = cross(pos, vel)

    inclination = np.arccos(
        specific_angular_momentum[:, 2] / norm(specific_angular_momentum, axis=1)
    )

    return inclination


def gas_mass(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the gas mass from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The gas mass on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        dustfrac: ndarray = snap['dustfrac'][h > 0]
        mass: ndarray = snap['mass'][h > 0]
    else:
        dustfrac = snap['dustfrac']
        mass = snap['mass']

    gasfrac = 1 - dustfrac.sum(axis=1)
    return gasfrac * mass


def dust_mass(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the dust mass per species from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The dust mass per species on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        dustfrac: ndarray = snap['dustfrac'][h > 0]
        mass: ndarray = snap['mass'][h > 0]
    else:
        dustfrac = snap['dustfrac']
        mass = snap['mass']

    return dustfrac * mass[:, np.newaxis]


def gas_density(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the gas density from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The gas density on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        dustfrac: ndarray = snap['dustfrac'][h > 0]
        density: ndarray = snap['density'][h > 0]
    else:
        dustfrac = snap['dustfrac']
        density = snap['density']

    gasfrac = 1 - dustfrac.sum(axis=1)
    return gasfrac * density


def dust_density(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the dust density per species from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The dust density per species on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        dustfrac: ndarray = snap['dustfrac'][h > 0]
        density: ndarray = snap['density'][h > 0]
    else:
        dustfrac = snap['dustfrac']
        density = snap['density']

    return dustfrac * density[:, np.newaxis]


def radial_distance(
    snap: SnapLike, coordinates: str = 'cylindrical', ignore_accreted: bool = False
) -> ndarray:
    """Calculate the radial distance.

    Can compute the radial distance in cylindrical and spherical
    coordinates.

    Parameters
    ----------
    snap
        The Snap object.
    coordinates : optional
        Either 'cylindrical' or 'spherical'. Default is 'cylindrical'.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The radial distance on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
    else:
        pos = snap['position']

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]

    if coordinates == 'cylindrical':
        return np.sqrt(x ** 2 + y ** 2)
    elif coordinates == 'spherical':
        return np.sqrt(x ** 2 + y ** 2 + z ** 2)
    else:
        raise ValueError('Cannot determine coordinates')


def azimuthal_angle(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the azimuthal angle.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
    else:
        pos = snap['position']

    x, y = pos[:, 0], pos[:, 1]
    return np.arctan2(y, x)


def polar_angle(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the polar angle.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
    else:
        pos = snap['position']

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    return np.arccos(z / r)


def radial_velocity(
    snap: SnapLike, coordinates: str = 'cylindrical', ignore_accreted: bool = False
) -> ndarray:
    """Calculate the radial velocity.

    Can compute the radial velocity in cylindrical and spherical
    coordinates.

    Parameters
    ----------
    snap
        The Snap object.
    coordinates : optional
        Either 'cylindrical' or 'spherical'. Default is 'cylindrical'.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The radial velocity on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    vx, vy, vz = vel[:, 0], vel[:, 1], vel[:, 2]

    if coordinates == 'cylindrical':
        vr = (x * vx + y * vy) / np.sqrt(x ** 2 + y ** 2)
    elif coordinates == 'spherical':
        vr = (x * vx + y * vy + z * vz) / np.sqrt(x ** 2 + y ** 2 + z ** 2)
    else:
        raise ValueError('Cannot determine coordinates')

    return vr


def angular_velocity(snap: SnapLike, ignore_accreted: bool = False) -> ndarray:
    """Calculate the angular velocity.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    ndarray
        The angular velocity on the particles.
    """
    if ignore_accreted:
        h: ndarray = snap['smooth']
        pos: ndarray = snap['position'][h > 0]
        vel: ndarray = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    x, y = pos[:, 0], pos[:, 1]
    vx, vy = vel[:, 0], vel[:, 1]

    vphi = (x * vy - y * vx) / (x ** 2 + y ** 2)

    return vphi



"""
Plonk
=====

**Plonk** is a Python package for analyzing and visualizing smoothed
particle hydrodynamics simulation data.

Features
--------

- Read in Phantom HDF snapshot files.
- Read in global quantity evolution files.
- Encapsulate entire simulation data as Simulation object.
- Access particle and sink arrays.
- Access simulation parameters and units.
- Compute extra quantities on particles.
- Visualize data using kernel interpolation.

Classes
-------

- Snap
    Represents a smoothed particle hydrodynamics snapshot file,
    containing particles, sinks, and file header information.

- Evolution
    Represents globally averaged quantities as time series data.

- Simulation
    Represents an entire smoothed particle hydrodynamics simulation.
    It contains instances of Snap and Evolution objects.

- Visualization
    Represents a visualization of a Snap object.

Subpackages
-----------

- analysis
    Contains classes and functions for performing analysis on snapshot
    files.

- simulation
    Contains classes and functions for accessing multiple simulation
    files as a coherent object.

- snap
    Contains classes and functions for reading and accessing snapshot
    files.

- utils
    Contains utility classes and functions.

- visualize
    Contains classes and functions for visualization of snapshot files.

Documentation
-------------

See https://plonk.readthedocs.io/ for documentation. The source code is
available at https://github.com/dmentipl/plonk.
"""

from typing import Any

import pint

units: Any = pint.UnitRegistry(system='cgs')
Quantity: Any = units.Quantity

from . import analysis, simulation, snap, utils, visualize
from .analysis import Profile
from .simulation import Evolution, Simulation, load_ev, load_sim
from .snap import Snap, load_snap
from .visualize import Visualization


__all__ = (
    ['Evolution', 'Profile', 'Simulation', 'Snap', 'Visualization']  # Classes
    + ['analysis', 'simulation', 'snap', 'utils', 'visualize']  # Packages
    + ['load_snap', 'load_ev', 'load_sim']  # User functions
    + ['units', 'Quantity']
)

# Canonical version number
__version__ = '0.4.0'

# Add units
units.define('solar_mass = 1.9891e33 g')
units.define('solar_radius = 6.959500e10 cm')
units.define('earth_mass = 5.979e27 g')
units.define('earth_radius = 6.371315e8 cm')
units.define('jupiter_mass = 1.89813e30 g')



"""Simulation class for smoothed particle hydrodynamics simulations.

This class contains all information associated with a simulation. It
contains the snapshot as a list of Snap objects, and the global quantity
and sink quantity time series files an Evolution objects.
"""

from __future__ import annotations

from copy import copy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from ..snap import Snap, load_snap
from .evolution import Evolution, load_ev

_properties_vary_per_snap = ('time',)


class Simulation:
    """Smoothed particle hydrodynamics simulation object.

    This class aggregates snapshot files, global quantity and sink time
    series data. Snapshot files contain a complete snapshot of the
    simulation at a particular time. Other files contain time series of
    global quantities on particles such as energy and momentum, and
    time series data for sink particles.

    Examples
    --------
    Reading simulation data into a Simulation object.

    >>> sim = plonk.load_sim('prefix', path_to_directory)

    Accessing the snapshots.

    >>> sim.snaps

    Accessing the global quantity and sink time series data.

    >>> sim.global_quantities
    >>> sim.sink_quantities
    """

    def __init__(self):

        self.prefix: str
        self.path: Path

        self._snaps: List[Snap] = None
        self._properties: Dict[str, Any] = None
        self._global_quantities: Evolution = None
        self._sink_quantities: List[Evolution] = None

        self._len = -1

    def load_sim(
        self, prefix: str, directory: Optional[Union[str, Path]] = None
    ) -> Simulation:
        """Load Simulation.

        Parameters
        ----------
        prefix
            Simulation prefix, e.g. 'disc', if files are named like
            disc_00000.h5, disc01.ev, discSink0001N01.ev, etc.
        directory
            Directory containing simulation snapshot files and auxiliary
            files.
        """
        if not isinstance(prefix, str):
            raise TypeError('prefix must be str')

        if directory is None:
            directory = '.'
        else:
            if not isinstance(directory, (str, Path)):
                raise TypeError('directory must be str or pathlib.Path')

        self.prefix = prefix
        self.path = Path(directory).expanduser().resolve()

        if not list(self.path.glob(self.prefix + '*')):
            raise FileNotFoundError(f'No files with prefix: {prefix}')

        self._snap_file_type, self._snap_file_extension = self._get_snap_file_type()

        self._snap_files = self._get_snap_files()
        self._global_ev_files = self._get_global_ev_files()
        self._sink_ev_files = self._get_sink_ev_files()

        return self

    @property
    def snaps(self) -> List[Snap]:
        """List of Snap objects associated with the simulation."""
        if self._snaps is None:
            self._generate_snap_objects()

        return self._snaps

    @property
    def properties(self) -> Dict[str, Any]:
        """Properties associated with the simulation."""
        if self._properties is None:
            self._generate_properties()

        return self._properties

    @property
    def global_quantities(self) -> Evolution:
        """Global quantity time series data."""
        if self._global_quantities is None:
            self._generate_global_quantities()

        return self._global_quantities

    @property
    def sink_quantities(self) -> List[Evolution]:
        """Sink time series data."""
        if self._sink_quantities is None:
            self._generate_sink_quantities()

        return self._sink_quantities

    def _generate_snap_objects(self):
        """Generate Snap objects."""
        snaps = list()
        for snap in self._snap_files:
            snaps.append(load_snap(snap))
        self._snaps = snaps

    def _generate_properties(self):
        """Generate sim.properties from snap.properties."""
        prop = copy(self.snaps[0].properties)
        for key in _properties_vary_per_snap:
            prop[key] = list()
        prop
        for snap in self.snaps:
            for key, val in snap.properties.items():
                if isinstance(prop[key], list):
                    prop[key].append(val)
                else:
                    if prop[key] != val:
                        prop[key] = '__inconsistent__'
        self._properties = prop

    def _generate_global_quantities(self):
        """Generate global quantity time series objects."""
        self._global_quantities = None
        if self._global_ev_files:
            self._global_quantities = load_ev(self._global_ev_files)

    def _generate_sink_quantities(self):
        """Generate sink quantity time series objects."""
        self._sink_quantities = list()
        [self._sink_quantities.append(load_ev(files)) for files in self._sink_ev_files]

    def _get_global_ev_files(self, glob: str = None) -> List[Path]:
        """Get global ev files."""
        if glob is None:
            # Phantom ev file name format
            glob = self.prefix + '[0-9][0-9].ev'

        return sorted(list(self.path.glob(glob)))

    def _get_sink_ev_files(self, glob: str = None) -> List[List[Path]]:
        """Get sink ev files."""
        if glob is None:
            # Phantom ev file name format
            glob = self.prefix + 'Sink[0-9][0-9][0-9][0-9]N[0-9][0-9].ev'

        n = len(self.prefix) + len('Sink')
        n_sinks = len(set([p.name[n : n + 4] for p in list(self.path.glob(glob))]))

        sinks = list()
        for idx in range(1, n_sinks + 1):
            sinks.append(
                sorted(
                    list(self.path.glob(self.prefix + f'Sink{idx:04}N[0-9][0-9].ev'))
                )
            )

        return sinks

    def _get_snap_files(self, glob: str = None) -> List[Path]:
        """Get snapshot files."""
        if glob is None:
            # Phantom snapshot file name format
            glob = (
                self.prefix + '_[0-9][0-9][0-9][0-9][0-9].' + self._snap_file_extension
            )

        return sorted(list(self.path.glob(glob)))

    def _get_snap_file_type(self, glob: str = None) -> Tuple[str, str]:
        """Snapshot file type.

        Determine snap file type from extension assuming file names
        follow a glob pattern.
        """
        if glob is None:
            # Phantom HDF5 snapshot file name format
            glob = self.prefix + '_[0-9][0-9][0-9][0-9][0-9].h5'

        file_types = set([f.suffix for f in self.path.glob(glob)])

        if len(file_types) > 1:
            raise ValueError(
                'Cannot determine simulation snapshot file type: '
                f'is it one of {file_types}?'
            )
        elif len(file_types) == 0:
            raise ValueError(
                'Cannot determine snapshot file type: '
                'no files named like prefix_xxxxx.ext'
            )

        file_ext = file_types.pop()[1:]

        if file_ext == 'h5':
            file_type = 'HDF5'

        return file_type, file_ext

    def __len__(self):
        """Length as number of snaps."""
        if self._len == -1:
            self._len = len(self.snaps)
        return self._len

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.Simulation: "{self.prefix}", ' f'directory="{self.path.name}">'


def load_sim(prefix: str, directory: Optional[Union[str, Path]] = None) -> Simulation:
    """Load Simulation.

    Parameters
    ----------
    prefix
        Simulation prefix, e.g. 'disc', if files are named like
        disc_00000.h5, disc01.ev, discSink0001N01.ev, etc.
    directory
        Directory containing simulation snapshot files and auxiliary
        files.
    """
    return Simulation().load_sim(prefix=prefix, directory=directory)



"""The visualization sub-package.

The Plonk implementation for visualizing smoothed particle hydrodynamics
simulations using kernel density estimation based interpolation.
"""

from .animation import animation, animation_profiles
from .interpolation import interpolate
from .visualization import Visualization, plot, plot_snaps, str_to_units

__all__ = [
    'Visualization',
    'animation',
    'animation_profiles',
    'interpolate',
    'plot',
    'plot_snaps',
    'str_to_units',
]



"""Phantom HDF5 snapshot file."""

from __future__ import annotations

from pathlib import Path
from typing import Callable, Union

import numpy as np
from numpy import ndarray

from ... import units
from ..snap import Snap
from ..units import generate_units_dictionary
from .hdf5 import HDF5File


class PhantomHDF5Snap:
    """Phantom HDF5 reader."""

    def __init__(self):
        self.snap = Snap()
        self.snap._file_pointer = None

    def generate_snap_from_file(self, filename: Union[str, Path]) -> Snap:
        """Generate a Snap object from a Phantom HDF5 file.

        Parameters
        ----------
        filename
            The path to the file.

        Returns
        -------
        Snap
            A Snap object.
        """
        self.hdf5_file = HDF5File(filename)
        self.snap._file_pointer = self.hdf5_file.file_handle

        self._header = {
            key: val[()] for key, val in self.hdf5_file.file_handle['header'].items()
        }
        self._header_to_properties(self._header)

        self._populate_array_registry()

        if self._header['nptmass'] > 0:
            self._populate_sink_arrays()

        return self.snap

    def _header_to_properties(self, header: dict):

        self.snap.properties['time'] = header['time']

        length = header['udist'] * units.cm
        time = header['utime'] * units.s
        mass = header['umass'] * units.g
        self.snap.units = generate_units_dictionary(length, mass, time)

        self.snap.properties['hfact'] = header['hfact']

        self.snap.properties['ieos'] = header['ieos']
        self.snap.properties['gamma'] = header['gamma']
        self.snap.properties['polyk'] = 2 / 3 * header['RK2']
        self.snap.properties['qfacdisc'] = header['qfacdisc']

        n_dust = header['ndustsmall'] + header['ndustlarge']
        if n_dust > 0:
            self.snap.properties['grain size'] = header['grainsize'][:n_dust]
            self.snap.properties['grain density'] = header['graindens'][:n_dust]

    def _populate_array_registry(self):

        arrays = list(self.hdf5_file.file_handle['particles'])

        self.snap._array_registry['position'] = _get_dataset('xyz', 'particles')
        self.snap._array_registry['smooth'] = _get_dataset('h', 'particles')

        if 'vxyz' in self.snap._file_pointer['particles']:
            self.snap._array_registry['velocity'] = _get_dataset('vxyz', 'particles')
            arrays.remove('vxyz')

        self.snap._array_registry['type'] = _particle_type
        if self._header['ndustlarge'] > 0:
            self.snap._array_registry['dust_type'] = _dust_particle_type

        self.snap._array_registry['mass'] = _mass
        self.snap._array_registry['density'] = _density

        for array in ('xyz', 'itype', 'h'):
            arrays.remove(array)

        for array in arrays:
            self.snap._array_registry[array] = _get_dataset(array, 'particles')

    def _populate_sink_arrays(self):

        name_map = {
            'xyz': 'position',
            'vxyz': 'velocity',
            'm': 'mass',
            'h': 'smooth',
            'hsoft': 'softening',
            'maccreted': 'maccreted',
            'spinxyz': 'spin',
            'tlast': 'tlast',
        }

        for name_on_file, name in name_map.items():
            self.snap._sink_registry[name] = _get_dataset(name_on_file, 'sinks')


def _get_dataset(dataset: str, group: str) -> Callable:
    def func(snap: Snap) -> ndarray:
        return snap._file_pointer[f'{group}/{dataset}'][()]

    return func


def _particle_type(snap: Snap) -> ndarray:
    idust = _get_dataset('idust', 'header')(snap)
    particle_type = np.abs(_get_dataset('itype', 'particles')(snap))
    particle_type[particle_type >= idust] = 2
    return particle_type


def _dust_particle_type(snap: Snap) -> ndarray:
    idust = _get_dataset('idust', 'header')(snap)
    particle_type = np.abs(_get_dataset('itype', 'particles')(snap))
    dust_type = np.zeros(particle_type.shape, dtype=np.int8)
    dust_type[particle_type >= idust] = (
        particle_type[particle_type >= idust] - idust + 1
    )
    return dust_type


def _mass(snap: Snap) -> ndarray:
    massoftype = snap._file_pointer['header/massoftype'][:]
    particle_type = _get_dataset('itype', 'particles')(snap)
    return massoftype[particle_type - 1]


def _density(snap: Snap) -> ndarray:
    m = _mass(snap)
    h = _get_dataset('h', 'particles')(snap)
    hfact = snap.properties['hfact']
    return m * (hfact / h) ** 3



"""Sink quantities.

Calculate various quantities relating to sink particles.
"""

import numpy as np


def Roche_sphere(m1: float, m2: float, separation: float):
    """Calculate an estimate of the Roche sphere.

    Uses the formula from Eggleton (1983) ApJ 268, 368-369.

    Parameters
    ----------
    m1
        The mass of the body around which to calculate the Roche sphere.
    m2
        The mass of the second body.
    separation
        The distance between the bodies.
    """
    q = m1 / m2
    return (
        separation
        * 0.49
        * q ** (2 / 3)
        / (0.6 * q ** (2 / 3) + np.log(1.0 + q ** (1 / 3)))
    )


def Hill_radius(M: float, m: float, semi_major_axis: float):
    """Calculate the Hill radius.

    This calculation assumes zero eccentricity.

    Parameters
    ----------
    M
        The mass of the heavier body.
    m
        The mass of the smaller body orbiting the heavy body.
    semi_major_axis
        The semi-major axis of the smaller body.
    """
    return semi_major_axis * (m / (3 * M)) ** (1 / 3)



"""Plonk setup.py."""

import io
import pathlib
import re

from setuptools import setup

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',  # It excludes inline comment too
    io.open('plonk/__init__.py', encoding='utf_8_sig').read(),
).group(1)

install_requires = [
    'h5py',
    'matplotlib',
    'numba',
    'numpy',
    'pandas',
    'pint',
    'scikit-image>=0.16',
    'scipy',
]
packages = [
    'plonk',
    'plonk.analysis',
    'plonk.simulation',
    'plonk.snap',
    'plonk.snap.readers',
    'plonk.utils',
    'plonk.visualize',
]

description = 'Smoothed particle hydrodynamics analysis and visualization with Python.'
long_description = (pathlib.Path(__file__).parent / 'README.md').read_text()

setup(
    name='plonk',
    version=__version__,
    author='Daniel Mentiplay',
    author_email='d.mentiplay@gmail.com',
    url='https://github.com/dmentipl/plonk',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=packages,
    license='MIT',
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



"""Animations of visualizations."""

import pathlib
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
from matplotlib import animation as _animation
from numpy import ndarray

from ..analysis import Profile
from ..snap import SnapLike
from ..utils import get_extent_from_percentile
from .visualization import interpolate, plot

_interp_kwargs = ('number_of_pixels', 'density_weighted')


def animation(
    *,
    filename: Union[str, Path],
    snaps: List[SnapLike],
    quantity: Optional[Union[str, ndarray]] = None,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    quantity
        The quantity to visualize. Can be a string to pass to Snap, or
        a 1d array (N,) of scalar data, or a 2d array (N, 3) of
        vector data. If quantity is 2d, only the first two components
        are interpolated, i.e. quantity[:, 0] and quantity[:, 1].
        Default is None.
    text
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.plot.
    """
    filepath = pathlib.Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    interp = kwargs.get('interp', 'projection')

    fig, ax = plt.subplots()
    viz = plot(snap=snaps[0], quantity=quantity, ax=ax, **kwargs)
    im = viz.objects['image']
    if im is None:
        raise NotImplementedError(
            'Can only currently produce animations of rendered images. (Or profiles\n'
            'with animation_profile.)'
        )
    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    def animate(idx):
        print(f'Visualizing snap: {idx}')
        _kwargs = {k: v for k, v in kwargs.items() if k in _interp_kwargs}
        extent = kwargs.get('extent', get_extent_from_percentile(snaps[idx]))
        interp_data = interpolate(
            snap=snaps[idx], interp=interp, quantity=quantity, extent=extent, **_kwargs,
        )
        im.set_data(interp_data)
        im.set_extent(extent)
        im.axes.set_xlim(extent[:2])
        im.axes.set_ylim(extent[2:])
        vmin = kwargs.get('vmin', interp_data.min())
        vmax = kwargs.get('vmax', interp_data.max())
        im.set_clim(vmin, vmax)
        if text is not None:
            _text.set_text(text[idx])
        return [im]

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(snaps), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()

    return anim


def animation_profiles(
    *,
    filename: Union[str, Path],
    profiles: List[Profile],
    quantity: str,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    profiles
        A list of Profile objects to animate.
    quantity
        The quantity to profile. Can be a string to pass to Profile.
    text
        List of strings to display per profile plot.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to plot function.
    """
    filepath = pathlib.Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    fig, ax = plt.subplots()
    [line] = ax.plot(profiles[0]['radius'], profiles[0][quantity], **kwargs)
    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    def animate(idx):
        print(f'Plotting profile: {idx}')
        line.set_data(profiles[idx]['radius'], profiles[idx][quantity])
        if text is not None:
            _text.set_text(text[idx])
        return [line]

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(profiles), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()

    return anim



"""The simulation sub-package.

It contains Plonk the implementation of smoothed particle hydrodynamics
combined simulation data. This includes "evolution" data.
"""

from .evolution import Evolution, load_ev
from .simulation import Simulation, load_sim

__all__ = ['Evolution', 'Simulation', 'load_ev', 'load_sim']



"""Evolution class for global quantites.

This module contains the Evolution class for tracking global quantities
and sink particle time series data. These files track averaged
quantities that are more frequently output than snapshot files.
"""

from __future__ import annotations

from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame, Series


class Evolution:
    """Smoothed particle hydrodynamics simulation time series object.

    Evolution files track global quantities, such as energy, momentum,
    and density, over time. The time increments in these files is
    smaller than the snapshot file output time. These files are
    typically stored as text files.

    The data is stored as a pandas DataFrame.

    Examples
    --------
    Reading a single evolution file into an Evolution object.

    >>> file_name = 'simulation.ev'
    >>> ev = plonk.load_ev(file_name)

    Reading a collection of evolution files into an Evolution object.

    >>> file_names = ('sim01.ev', 'sim02.ev', 'sim03.ev')
    >>> ev = plonk.load_ev(file_names)

    See what columns and what data are available.

    >>> ev.columns
    >>> ev.data

    Accessing part of the data as a pandas DataFrame or Series.

    >>> time = ev['time']
    >>> first_100 = ev[:100]

    Plot columns using pandas plotting methods.

    >>> ev.plot('time', ['x', 'y'])
    """

    def __init__(self):

        self.file_paths: Tuple[Path, ...]
        self.file_names: Tuple[str, ...]

    def load_from_file(
        self,
        filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
    ) -> Evolution:
        """Load from file(s).

        Parameters
        ----------
        filename(s)
            Collection of paths to evolution file(s) in chronological order.
            These should all contain the same columns.
        """
        if isinstance(filenames, (str, Path)):
            _filenames = [filenames]
        elif isinstance(filenames, (list, tuple)):
            _filenames = list(filenames)
        else:
            raise ValueError('filenames is not a known type')

        _file_paths = list()
        _file_names = list()
        for filename in _filenames:
            path = Path(filename)
            _file_paths.append(path.resolve())
            _file_names.append(path.name)

        self.file_paths = tuple(_file_paths)
        self.file_names = tuple(_file_names)
        _check_file_consistency(self.file_paths)

        self._columns = _get_columns(self.file_paths[0])
        self._data = self._get_data()

        return self

    @property
    def columns(self) -> Tuple[str, ...]:
        """List of available time evolution data."""
        return self._columns

    @property
    def data(self) -> DataFrame:
        """Time evolution data as a pandas DataFrame."""
        return self._data

    def plot(self, *args, **kwargs):
        """Plot using pandas."""
        return self.data.plot(*args, **kwargs)

    def __getitem__(
        self, inp: Union[str, List[str], ndarray, int, slice]
    ) -> Union[DataFrame, Series]:
        """Return a pandas object."""
        if isinstance(inp, str):
            return self.data[inp]
        elif isinstance(inp, (ndarray, int, slice)):
            return self.data.iloc[inp]
        elif isinstance(inp, list):
            if isinstance(inp[0], str):
                return self.data[inp]
            elif isinstance(inp[0], int):
                return self.data.iloc[inp]
        raise ValueError('Cannot determine item to return')

    def __setitem__(self, name: str, item: Union[ndarray, Series]):
        """Set a column with ndarray or pandas Series."""
        if not isinstance(item, (ndarray, Series)):
            raise ValueError('"item" must be NumPy ndarray or pandas Series')
        if item.shape[0] != len(self):
            raise ValueError('Length of array does not match length')
        if name in self.columns:
            raise ValueError('Attempting overwrite existing column')
        else:
            self.data[name] = item

    def _get_data(self) -> DataFrame:

        times = list()
        for filename in self.file_paths:
            times.append(np.loadtxt(filename, usecols=0))

        _skiprows = [0]
        if len(times) > 1:
            for t1, t2 in zip(times, times[1:]):
                _skiprows.append(np.where(t2 < t1[-1])[0][-1] + 2)

        df = pd.concat(
            (
                pd.read_csv(
                    f,
                    names=self._columns,
                    skiprows=skiprows,
                    skipinitialspace=True,
                    delim_whitespace=True,
                    comment='#',
                )
                for f, skiprows in zip(self.file_paths, _skiprows)
            )
        )

        df.reset_index(inplace=True, drop=True)
        return df

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.Evolution: "{self.file_names}">'

    def __len__(self):
        """Dunder len method."""
        return len(self.data)


def _get_columns(filename: Path) -> Tuple[str, ...]:

    with open(filename) as f:
        column_line = f.readline().strip('\n')

    _column_line = [item.strip('] ')[2:].strip(' ') for item in column_line.split('[')]

    return tuple(_column_line[1:])


def _check_file_consistency(filenames: Tuple[Path, ...]) -> None:

    columns = _get_columns(filenames[0])
    for filename in filenames:
        columns_previous = columns
        columns = _get_columns(filename)
        if columns != columns_previous:
            raise ValueError('files have different columns')


def load_ev(
    filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
) -> Evolution:
    """Load Evolution from file(s).

    Parameters
    ----------
    filename(s)
        Collection of paths to evolution file(s) in chronological order.
        These should all contain the same columns.
    """
    return Evolution().load_from_file(filenames)
